#
# This implementation is derived from the ``concurrent.futures``
# module of Python 3.2, by Brian Quinlan, (C) 2011 the Python Software
# Foundation. See http://docs.python.org/3/license.html for more information.

import logging
import signal
import threading
import uuid
from itertools import islice
from contextlib import contextmanager

import h5py


log = logging.getLogger(__name__)


class WorkManager:
    '''Base class for all work managers. At a minimum, work managers must provide a
    ``submit()`` function and a ``n_workers`` attribute (which may be a property),
    though most will also override ``startup()`` and ``shutdown()``.'''

    @classmethod
    def from_environ(cls, wmenv=None):
        raise NotImplementedError

    @classmethod
    def add_wm_args(cls, parser, wmenv=None):
        return

    def __repr__(self):
        return '<{classname} at 0x{id:x}>'.format(classname=self.__class__.__name__, id=id(self))

    def __init__(self):
        self._sigint_handler_installed = False
        self.prior_sigint_handler = None
        self.running = False

    def __enter__(self):
        self.startup()
        return self

    def __exit__(self, exc_type, exc_val, exc_traceback):
        self.shutdown()
        return False

    def sigint_handler(self, signum, frame):
        self.shutdown()
        if self.prior_sigint_handler in (signal.SIG_IGN, None):
            pass
        elif self.prior_sigint_handler == signal.SIG_DFL:
            raise KeyboardInterrupt
        else:
            self.prior_sigint_handler(signum, frame)

    def install_sigint_handler(self):
        if not self._sigint_handler_installed:
            self._sigint_handler_installed = True
            self.prior_sigint_handler = signal.signal(signal.SIGINT, self.sigint_handler)

    def startup(self):
        '''Perform any necessary startup work, such as spawning clients.'''
        self.running = True

    def shutdown(self):
        '''Cleanly shut down any active workers.'''
        self.running = False

    def run(self):
        '''Run the worker loop (in clients only).'''
        pass

    def submit(self, fn, args=None, kwargs=None):
        '''Submit a task to the work manager, returning a `WMFuture` object representing the pending
        result. ``fn(*args,**kwargs)`` will be executed by a worker, and the return value assigned as the
        result of the returned future.  The function ``fn`` and all arguments must be picklable; note
        particularly that off-path modules (like the system module and any active plugins) are not
        picklable unless pre-loaded in the worker process (i.e. prior to forking the master).'''
        raise NotImplementedError

    def submit_many(self, tasks):
        '''Submit a set of tasks to the work manager, returning a list of `WMFuture` objects representing
        pending results. Each entry in ``tasks`` should be a triple (fn, args, kwargs), which will result in
        fn(*args, **kwargs) being executed by a worker. The function ``fn`` and all arguments must be
        picklable; note particularly that off-path modules are not picklable unless pre-loaded in the worker
        process.'''

        return [self.submit(fn, args, kwargs) for (fn, args, kwargs) in tasks]

    def as_completed(self, futures):
        '''Return a generator which yields results from the given ``futures`` as they become
        available.'''
        pending = set(futures)

        # See which futures have results, and install a watcher on those that do not
        with WMFuture.all_acquired(pending):
            completed = {future for future in futures if future.done}
            pending -= completed
            watcher = FutureWatcher(pending, threshold=1)

        # Yield available results immediately
        for future in completed:
            yield future
        del completed

        # Wait on any remaining results
        while pending:
            watcher.wait()
            completed = watcher.reset()
            for future in completed:
                yield future
                pending.remove(future)

    def submit_as_completed(self, task_generator, queue_size=None):
        '''Return a generator which yields results from a set of ``futures`` as they become
        available. Futures are generated by the ``task_generator``, which must return a triple of the form
        expected by ``submit``. The method also accepts an int ``queue_size`` that dictates the
        maximum number of Futures that should be pending at any given time. The default value of
        ``None`` submits all of the tasks at once.'''

        futures = [self.submit(fn, args, kwargs) for (fn, args, kwargs) in islice(task_generator, queue_size)]
        pending = set(futures)

        with WMFuture.all_acquired(pending):
            watcher = FutureWatcher(pending, threshold=1)

        while pending:
            watcher.wait()
            completed = watcher.reset()
            new_futures = [self.submit(fn, args, kwargs) for (fn, args, kwargs) in islice(task_generator, len(completed))]
            pending.update(new_futures)

            with WMFuture.all_acquired(new_futures):
                watcher.add(new_futures)

            for future in completed:
                yield future
                pending.remove(future)

    def wait_any(self, futures):
        '''Wait on any of the given ``futures`` and return the first one which has a result available.
        If more than one result is or becomes available simultaneously, any completed future may be returned.'''
        pending = set(futures)
        with WMFuture.all_acquired(pending):
            completed = {future for future in futures if future.done}

            if completed:
                # If any futures are complete, then we don't need to do anything else
                return completed.pop()
            else:
                # Otherwise, we need to install a watcher
                watcher = FutureWatcher(futures, threshold=1)

        watcher.wait()
        completed = watcher.reset()
        return completed.pop()

    def wait_all(self, futures):
        '''A convenience function which waits on all the given ``futures`` in order.  This function returns
        the same ``futures`` as submitted to the function as a list, indicating the order in which waits
        occurred.'''
        futures = list(futures)
        results = []
        for future in futures:
            results.append(future.result)
        return futures

    @property
    def is_master(self):
        '''True if this is the master process for task distribution. This is necessary, e.g., for
        MPI, where all processes start identically and then must branch depending on rank.'''
        return True


class FutureWatcher:
    '''A device to wait on multiple results and/or exceptions with only one lock.'''

    def __init__(self, futures, threshold=1):
        self.event = threading.Event()
        self.lock = threading.RLock()
        self.threshold = threshold
        self.completed = []

        for future in futures:
            future._add_watcher(self)

    def signal(self, future):
        '''Signal this watcher that the given future has results available. If this
        brings the number of available futures above signal_threshold, this watcher's
        event object will be signalled as well.'''
        with self.lock:
            self.completed.append(future)
            if len(self.completed) >= self.threshold:
                self.event.set()

    def wait(self):
        '''Wait on one or more futures.'''
        return self.event.wait()

    def reset(self):
        '''Reset this watcher's list of completed futures, returning the list of completed futures
        prior to resetting it.'''
        with self.lock:
            self.event.clear()
            completed = self.completed
            self.completed = []
            return completed

    def add(self, futures):
        '''Add watchers to all futures in the iterable of futures.'''
        for future in futures:
            future._add_watcher(self)


class WMFuture:
    '''A "future", representing work which has been dispatched for completion asynchronously.'''

    @staticmethod
    @contextmanager
    def all_acquired(futures):
        '''Context manager to acquire all locks on the given ``futures``. Primarily for internal use.'''
        futures = list(futures)
        for future in futures:
            future._condition.acquire()

        yield  # to contents of "with" block

        for future in futures:
            future._condition.release()

    def __init__(self, task_id=None):
        self.task_id = task_id or uuid.uuid4()

        self._condition = threading.Condition()
        self._done = False
        self._result = None
        self._exception = None
        self._traceback = None

        # a set of Events representing who is waiting on results from this future
        # this set will be cleared after the result is updated and watchers are notified
        self._watchers = set()

        # a set of functions that will be called with this future as an argument when it is updated with a
        # result. This list will be cleared after the result is updated and all callbacks invoked
        self._update_callbacks = []

    def __repr__(self):
        return '<WMFuture 0x{id:x}: {self.task_id!s}>'.format(id=id(self), self=self)

    def __hash__(self):
        return hash(self.task_id)

    def _notify_watchers(self):
        '''Notify all watchers that this future has been updated, then deletes the list of update watchers.'''
        with self._condition:
            assert self._done
            for watcher in self._watchers:
                watcher.signal(self)
            self._watchers.clear()

    def _invoke_callbacks(self):
        '''Invoke all callbacks which have been registered on this future. Exceptions in callbacks
        will be logged and ignored.'''
        with self._condition:
            for callback in self._update_callbacks:
                try:
                    callback(self)
                except Exception:
                    # This may need to be a simple print to stderr, depending on the locking
                    # semantics of the logger.
                    log.exception('ignoring exception in result callback')
            del self._update_callbacks
            self._update_callbacks = []

    def _add_watcher(self, watcher):
        '''Add the given update watcher  to the internal list of watchers. If a result is available,
        returns immediately without updating the list of watchers.'''
        with self._condition:
            if self._done:
                watcher.signal(self)
                return
            else:
                self._watchers.add(watcher)

    def _add_callback(self, callback):
        '''Add the given update callback to the internal list of callbacks. If a result is available,
        invokes the callback immediately without updating the list of callbacks.'''
        with self._condition:
            if self._done:
                try:
                    callback(self)
                except Exception:
                    log.exception('ignoring exception in result callback')
            else:
                self._update_callbacks.append(callback)

    def _set_result(self, result):
        '''Set the result of this future to the given value, invoke on-completion callbacks, and notify
        watchers.'''
        with self._condition:
            self._result = result
            self._done = True
            self._condition.notify_all()
            self._invoke_callbacks()
            self._notify_watchers()

    def _set_exception(self, exception, traceback=None):
        '''Set the exception of this future to the given value, invoke on-completion callbacks, and notify
        watchers.'''

        with self._condition:
            self._exception = exception
            self._traceback = traceback
            self._done = True
            self._condition.notify_all()
            self._invoke_callbacks()
            self._notify_watchers()

    def get_result(self, discard=True):
        '''Get the result associated with this future, blocking until it is available.
        If ``discard`` is true, then removes the reference to the result contained
        in this instance, so that a collection of futures need not turn into a cache of
        all associated results.'''
        with self._condition:
            if self._done:
                if self._exception:
                    if isinstance(self._traceback, h5py.string_dtype(encoding='utf-8').type) or isinstance(self._traceback, str):
                        if self._traceback:
                            log.error('uncaught exception in remote function\n{}'.format(self._traceback))
                        raise self._exception
                    else:
                        raise self._exception.with_traceback(self._traceback)
            else:
                self._condition.wait()
                assert self._done
                if self._exception:
                    if isinstance(self._traceback, str):
                        log.error('uncaught exception in remote function\n{}'.format(self._traceback))
                        raise self._exception
                    else:
                        raise self._exception.with_traceback(self._traceback)

            result = self._result
            if discard:
                del self._result
            return result

    @property
    def result(self):
        return self.get_result(discard=False)

    def wait(self):
        '''Wait until this future has a result or exception available.'''
        with self._condition:
            if self._done:
                return
            else:
                self._condition.wait()
                assert self._done
                return

    def get_exception(self):
        '''Get the exception associated with this future, blocking until it is available.'''
        with self._condition:
            if self._done:
                return self._exception
            else:
                self._condition.wait()
                assert self._done
                return self._exception

    exception = property(get_exception, None, None, get_exception.__doc__)

    def get_traceback(self):
        '''Get the traceback object associated with this future, if any.'''
        with self._condition:
            if self._returned:
                return self._traceback
            else:
                self._condition.wait()
                assert self._done
                return self._traceback

    traceback = property(get_traceback, None, None, get_traceback.__doc__)

    def is_done(self):
        'Indicates whether this future is done executing (may block if this future is being updated).'
        with self._condition:
            return self._done

    done = property(is_done, None, None, is_done.__doc__)


# end class WMFuture



from __future__ import division, print_function; __metaclass__ = type

cimport cython
import numpy
cimport numpy
from numpy cimport uint32_t, uint64_t, int64_t, float64_t


_tt_dtype = numpy.dtype([('n_iter', numpy.uint32),
                         ('seg_id', numpy.int64),
                         ('parent_id', numpy.int64),
                         ('parent_offset', numpy.int64),  # offset of parent segment into this table
                         ('weight', numpy.float64),       # weight of this segment
                         ])

cdef packed struct _tt_rec:
    uint32_t  n_iter
    int64_t   seg_id
    int64_t   parent_id
    int64_t   parent_offset
    float64_t weight

cdef int64_t NO_PARENT = -1  #indicates that no parent is contained in the table; segment is a root
cdef uint32_t CT_CHUNKSIZE = 512
cdef uint32_t NODE_CHUNKSIZE = 512

cdef class _trajtree_base:
    cdef public object trajtable    # numpy.ndarray[_tt_rec, ndim=1]
    cdef public object childtable   # numpy.ndarray[object, ndim=1]
    cdef public object iter_offsets # dict

    #@cython.boundscheck(False)
    #@cython.wraparound(False)
    cpdef _build_table(self, segsel, data_manager):

        cdef uint32_t n_iter
        cdef uint32_t start_iter = segsel.start_iter
        cdef uint32_t stop_iter = segsel.stop_iter
        cdef uint64_t seg_count = len(segsel)

        cdef numpy.ndarray[_tt_rec, ndim=1] trajtable = numpy.empty((seg_count,), dtype=_tt_dtype)
        cdef numpy.ndarray[object, ndim=1]  childtable = numpy.empty((seg_count,), dtype=numpy.object_)
        cdef numpy.ndarray[int64_t, ndim=1] seg_ids = None
        cdef numpy.ndarray[int64_t, ndim=1] all_parent_ids = None
        cdef numpy.ndarray[float64_t, ndim=1] weights = None

        cdef uint64_t tt_offset = 0
        cdef int64_t seg_id = 0
        cdef int64_t parent_id = 0
        cdef int64_t parent_offset = NO_PARENT

        last_iter_indices = {} # mapping of seg_id -> table row for that seg_id in previous iteration
        iter_offsets = {}
        for n_iter in range(start_iter, stop_iter):
            iter_offsets[n_iter] = tt_offset
            this_iter_indices = {}

            seg_ids = numpy.fromiter(segsel.from_iter(n_iter), dtype=numpy.int64)
            iter_group = data_manager.get_iter_group(n_iter)
            seg_index = iter_group['seg_index']
            weights = seg_index['weight']
            all_parent_ids = numpy.array(data_manager.get_all_parent_ids(n_iter), numpy.int64)

            for seg_id in seg_ids:
                this_iter_indices[seg_id] = tt_offset

                # Record information about this segment
                trajtable[tt_offset].n_iter = n_iter
                trajtable[tt_offset].seg_id = seg_id
                trajtable[tt_offset].weight = weights[seg_id]
                trajtable[tt_offset].parent_id = all_parent_ids[seg_id]
                childtable[tt_offset] = None

                # Now record information about this segment's parent
                if n_iter == start_iter or trajtable[tt_offset].parent_id < 0:
                    trajtable[tt_offset].parent_offset = NO_PARENT
                else:
                    # where in the table is the parent?
                    # note that this will raise KeyError if a segment required for proper connectivity is
                    # missing from the input segment set
                    parent_offset = last_iter_indices[trajtable[tt_offset].parent_id]
                    #assert parent_offset != -1 and parent_offset < tt_offset
                    #assert trajtable[parent_offset].seg_id == trajtable[tt_offset].parent_id
                    #assert trajtable[parent_offset].n_iter == n_iter-1
                    trajtable[tt_offset].parent_offset = parent_offset

                    # all first-iteration segments are roots for the purposes of analysis
                    # as are true roots (from recycling)
                    if childtable[parent_offset] is None:
                        childtable[parent_offset] = [tt_offset]
                    else:
                        childtable[parent_offset].append(tt_offset)
                tt_offset += 1

            del last_iter_indices, seg_index
            last_iter_indices = this_iter_indices

        for tt_offset in range(len(trajtable)):
            if childtable[tt_offset] is not None:
                childtable[tt_offset] = numpy.array(childtable[tt_offset], dtype=numpy.int64)

        self.childtable = childtable
        self.trajtable = trajtable
        self.iter_offsets = iter_offsets

    cpdef uint64_t count_roots(self):
        cdef uint64_t i=0, root_count=0

        cdef numpy.ndarray[_tt_rec, ndim=1] trajtable = self.trajtable

        for i in range(0, len(trajtable)):
            if trajtable[i].parent_offset == NO_PARENT:
                root_count += 1
        return root_count

    cpdef uint64_t count_leaves(self):
        cdef uint64_t i=0, leaf_count=0

        cdef numpy.ndarray[object, ndim=1] childtable = self.childtable
        for i in range(0, len(childtable)):
            if childtable[i] is None:
                leaf_count += 1
        return leaf_count

    cpdef get_child_indices(self, int64_t index):
        child_indices = self.childtable[index]
        if child_indices is None:
            return []
        else:
            return child_indices



import collections
import threading
import time

import blessings  # @UnresolvedImport

import numpy as np
from scipy.stats import linregress


def nop():
    pass


class ProgressIndicator:
    def __init__(self, stream=None, interval=1):
        self.terminal = blessings.Terminal(stream=stream)
        self.interval = interval  # how often the status is updated

        self._operation = None  # string describing operation
        self._extent = None  # how far we have to go, total
        self._progress = 0  # how far we've gone

        self._endloop = False
        self._startup_event = threading.Event()
        self._event = threading.Event()

        self._last_update = None
        self._last_operation = None
        self._operation_start = None
        self._progress_history = collections.deque(maxlen=100)

        try:
            self.flush_output = self.terminal.stream.flush
        except AttributeError:
            self.flush_output = nop

        self.fancy = self.terminal.is_a_tty

    def _completion_est(self):
        if self._extent is None:
            return 'unknown'

        history = np.array(self._progress_history)
        if not len(history):
            return 'unknown'
        history[:, 1] -= self._extent
        (_slope, intercept, _r, _p, _stderr) = linregress(history[:, 1], history[:, 0])
        if not np.isfinite(intercept):
            return 'unknown'
        else:
            total_seconds_remaining = intercept - time.time()

            if total_seconds_remaining < 60:
                return 'less than 1 minute'

            # 1 minute or more remains
            minutes = total_seconds_remaining / 60
            if minutes < 60:
                minutes = int(round(minutes))
                return 'about {:d} {}'.format(minutes, 'minutes' if minutes > 1 else 'minute')

            # 1 hour or more remains
            hours, minutes = divmod(minutes, 60)
            days, hours = divmod(hours, 24)
            days = int(days)
            hours = int(round(hours))
            minutes = int(round(minutes))
            if days > 0:
                return 'about {:d} {}, {:d} {}'.format(days, 'days' if days > 1 else 'day', hours, 'hours' if hours > 1 else 'hour')
            else:
                return 'about {:d} {}, {:d} {}'.format(
                    hours, 'hours' if hours > 1 else 'hour', minutes, 'minutes' if minutes > 1 else 'minute'
                )

    def draw_fancy(self):
        operation_text = 'Operation:      '
        progress_text = 'Progress:       '
        remaining_text = 'Time remaining: '

        term = self.terminal
        stream = self.terminal.stream

        stream.write('{t.clear_eol}{t.bold}{}{t.normal}{}\n'.format(operation_text, self._operation or '', t=term))

        if self._extent:
            width = term.width
            pct_done = self.progress / self._extent
            pct_part = '{:<4.0%} '.format(pct_done)

            barwidth = width - len(progress_text) - 2 - len(pct_part)
            neqs = int(round(pct_done * barwidth))
            eqs = '=' * neqs
            spaces = ' ' * (barwidth - neqs)

            stream.write(
                '{t.clear_eol}{t.bold}{progress_text}{t.normal}{pct_part}{t.bold}[{t.normal}{eqs}{spaces}{t.bold}]{t.normal}\n'.format(
                    t=term, progress_text=progress_text, pct_part=pct_part, eqs=eqs, spaces=spaces
                )
            )

            completion_est = self._completion_est()
            stream.write('{t.clear_eol}{t.bold}{}{t.normal}{}\n'.format(remaining_text, completion_est, t=term))
            stream.write('{t.move_up}'.format(t=term) * 2)
        stream.write('{t.move_up}'.format(t=term))

        self._last_update = time.time()

    def draw_simple(self):
        if self._operation != self._last_operation:
            self.terminal.stream.write((self._operation or '(unknown)') + '...\n')
            self._last_operation = self._operation
            self._last_update = time.time()

    def draw(self):
        if not self._operation:
            return
        if self.fancy:
            self.draw_fancy()
        else:
            self.draw_simple()

    def clear(self):
        if self.fancy:
            if self._extent:
                nlines = 3
            else:
                nlines = 1

            for _i in range(nlines):
                self.terminal.stream.write('{t.clear_eol}{t.move_down}'.format(t=self.terminal))
            for _i in range(nlines):
                self.terminal.stream.write('{t.move_up}'.format(t=self.terminal))

    @property
    def operation(self):
        return self._operation

    @operation.setter
    def operation(self, op):
        if self._operation is not None:
            self.clear()
        self._operation = op
        self._operation_start = time.time()
        self._progress_history.clear()
        self._event.set()

    @property
    def extent(self):
        return self._extent

    @extent.setter
    def extent(self, ext):
        self._extent = ext
        self._progress_history.clear()
        self._event.set()

    @property
    def progress(self):
        return self._progress

    @progress.setter
    def progress(self, p):
        self._progress = p
        self._progress_history.append((time.time(), p))

    def new_operation(self, operation, extent=None, progress=0):
        self.operation = operation
        self.progress = progress
        self._extent = extent
        self._event.set()

    def _reporter_loop(self):
        self._startup_event.set()
        while not self._endloop:
            self._event.wait(self.interval)
            self._event.clear()
            if self._last_operation != self._operation or time.time() - self._last_update >= self.interval:
                self.draw()
                self.flush_output()

    def start(self):
        if self.fancy:
            self.terminal.stream.write(self.terminal.hide_cursor)
        self._endloop = False
        t = threading.Thread(target=self._reporter_loop)
        t.daemon = True
        t.start()
        self._startup_event.wait()
        self.flush_output()

    def stop(self):
        self._endloop = True
        self._event.set()

        self.clear()
        if self.fancy:
            self.terminal.stream.write(self.terminal.normal_cursor)
        self.flush_output()

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()
        return False


if __name__ == '__main__':
    with ProgressIndicator() as pi:
        pi.operation = 'Test 1'
        pi.extent = 10
        for i in range(10):
            pi.progress = i + 1
            time.sleep(2)

        time.sleep(0.2)



import numpy


class EDF:
    '''A class for creating and manipulating empirical distribution functions (cumulative
    distribution functions derived from sample data).
    '''

    @staticmethod
    def from_array(array):
        edf = EDF(None, None)
        edf.x = array[:, 0]
        edf.F = array[:, 1]
        edf.dF = numpy.diff(edf.F)
        return edf

    @staticmethod
    def from_arrays(x, F):
        edf = EDF(None, None)
        edf.x = x
        edf.F = F
        edf.dF = numpy.diff(edf.F)
        return edf

    def __init__(self, values, weights=None):
        '''Construct a new EDF from the given values and (optionally) weights.'''

        if values is None:
            self.x = None
            self.F = None
            self.dF = None
            return

        if weights is None:
            weights = numpy.ones((len(values)), numpy.float64)
        elif numpy.isscalar(weights):
            tweights = numpy.empty((len(values)), numpy.float64)
            tweights[:] = weights
            weights = tweights
        else:
            if len(weights) != len(values):
                raise TypeError('values and weights have different lengths')

        # Sort values
        sort_indices = numpy.argsort(values)
        values = values[sort_indices]
        weights = weights[sort_indices]

        # Determine unique abcissae; this is essentially stolen from numpy.lib.arraysetops.unique()
        x = values[numpy.concatenate(([True], values[1:] != values[:-1]))]
        F = numpy.empty((len(x),), numpy.float64)

        # ``values`` is arranged in increasing order, so we can walk along it and add up weights
        # as we go
        ival_last = 0
        ival = 0
        for ibin in range(0, len(x)):
            while ival < len(values) and values[ival] <= x[ibin]:
                ival += 1
            F[ibin] = weights[ival_last:ival].sum()
            ival_last = ival
        F = numpy.add.accumulate(F)
        F /= F[-1]

        self.x = x
        self.F = F
        self.dF = numpy.diff(F)

    def __len__(self):
        return len(self.x)

    def __call__(self, x):
        '''Evaluate this EDF at the given abcissae.'''
        indices = numpy.digitize(x, self.x)
        indices[indices >= len(self.x)] = len(self.x) - 1
        return self.F[indices]

    def as_array(self):
        '''Return this EDF as a (N,2) array, where N is the number of unique values passed to
        the constructor.  Numpy type casting rules are applied (so, for instance, integral abcissae
        are converted to floating-point values).'''

        result = numpy.empty((len(self.F), 2), dtype=numpy.result_type(self.x, self.F))
        result[:, 0] = self.x
        result[:, 1] = self.F
        return result

    def quantiles(self, p):
        '''Treating the EDF as a quantile function, return the values of the (statistical) variable whose
        probabilities are at least p.  That is, Q(p) = inf {x: p <= F(x) }.'''

        indices = numpy.searchsorted(self.F, p)
        indices[indices >= len(self.x)] = len(self.x) - 1
        return self.x[indices]

    def quantile(self, p):
        return self.quantiles([p])[0]

    def median(self):
        return self.quantiles([0.5])[0]

    def moment(self, n):
        '''Calculate the nth moment of this probability distribution

        <x^n> = int_{-inf}^{inf} x^n dF(x)
        '''

        if n == 1:
            return (self.x[:-1] * self.dF).sum()
        else:
            return (self.x[:-1] ** n * self.dF).sum()

    def cmoment(self, n):
        '''Calculate the nth central moment of this probability distribution'''

        if n < 2:
            return 0
        return ((self.x[:-1] - self.moment(1)) ** n * self.dF).sum()

    def mean(self):
        return self.moment(1)

    def var(self):
        '''Return the second central moment of this probability distribution.'''
        return self.cmoment(2)

    def std(self):
        '''Return the standard deviation (root of the variance) of this probability distribution.'''
        return self.cmoment(2) ** 0.5



import h5py
import numpy as np
from scipy.signal import fftconvolve
from warnings import warn

import westpa

from westpa.core.data_manager import weight_dtype, n_iter_dtype, vstr_dtype
from westpa.core.we_driver import NewWeightEntry
from westpa.core import h5io
from westpa.tools import WESTTool, WESTDataReader, IterRangeSelection
from westpa.tools.dtypes import iter_block_ci_dtype as ci_dtype
import westpa.mclib as mclib

fluxentry_dtype = np.dtype([('n_iter', n_iter_dtype), ('flux', weight_dtype), ('count', np.uint)])

target_index_dtype = np.dtype(
    [
        ('target_label', vstr_dtype),
        ('mean_flux', weight_dtype),
        ('mean_flux_ci_lb', weight_dtype),
        ('mean_flux_ci_ub', weight_dtype),
        ('mean_flux_correl_len', np.uintc),
    ]
)


def _extract_fluxes_fileversion_lt_7(iter_start, iter_stop, data_manager):
    '''Extract fluxes from old format, where groups for iterations where recyling
    occurs contain a 'recycling' table.'''

    assert data_manager.we_h5file_version < 7

    iter_count = iter_stop - iter_start
    target_count = data_manager.get_iter_group(iter_start)['recycling'].shape[0]
    fluxdata = np.zeros((iter_count,), dtype=fluxentry_dtype)

    if data_manager.we_h5file_version < 5:
        flux_field = 'weight'
    else:
        flux_field = 'flux'

    fluxdata = {itarget: np.zeros((iter_count,), dtype=fluxentry_dtype) for itarget in range(target_count)}

    for iiter, n_iter in enumerate(range(iter_start, iter_stop)):
        rdata = data_manager.get_iter_group(n_iter)['recycling']
        for itarget in range(target_count):
            fluxdata[itarget][iiter]['n_iter'] = n_iter
            fluxdata[itarget][iiter]['flux'] = rdata[itarget][flux_field]
            fluxdata[itarget][iiter]['count'] = rdata[itarget]['count']
        del rdata

    return fluxdata


def _extract_fluxes_fileversion_7(iter_start, iter_stop, data_manager):
    '''Extract fluxes from HDF5 file version 7, where recycling information is
    stored in the "new_weights" group of the iteration *following* recycling
    events.'''

    assert data_manager.we_h5file_version >= 7

    iter_count = iter_stop - iter_start
    iters = np.arange(iter_start, iter_stop, dtype=n_iter_dtype)

    # for each target by name, collect the iterations, fluxes, and counts
    # This is not the most foolproof way to do this, but it's good enough, and fast.
    # The most correct way to do this is tracing trajectories,
    # and warning if the boundary conditions change during the trace,
    # but that's for another tool.
    by_target = {}

    for iiter, n_iter in enumerate(range(iter_start, iter_stop)):
        target_states = data_manager.get_target_states(n_iter)
        try:
            new_weight_index = data_manager.get_iter_group(n_iter + 1)['new_weights']['index']
        except KeyError:
            # no recycling data available
            continue

        for tstate in target_states:
            try:
                target_info = by_target[tstate.label]
            except KeyError:
                # State not seen before
                target_info = by_target[tstate.label] = np.zeros((iter_count,), dtype=fluxentry_dtype)
                # If the target happens not to exist in an iteration (for whatever reason),
                # store a count of -1 as a sentinel
                target_info['count'][:] = -1
                target_info['n_iter'][:] = iters[:]

            recycled_from_tstate = (new_weight_index['source_type'] == NewWeightEntry.NW_SOURCE_RECYCLED) & (
                new_weight_index['target_state_id'] == tstate.state_id
            )

            recycle_count = recycled_from_tstate.sum()
            target_info['count'][iiter] = recycle_count
            if recycle_count:
                # flux is in units of per tau
                target_info['flux'][iiter] = new_weight_index[recycled_from_tstate]['weight'].sum()

        del new_weight_index, target_states

    # Find the last contiguous run where target is available
    for target_label in by_target:
        fluxdata = by_target[target_label]
        by_target[target_label] = fluxdata[np.searchsorted(fluxdata['count'], [0])[0] :]

    return by_target


def extract_fluxes(iter_start=None, iter_stop=None, data_manager=None):
    '''Extract flux values from the WEST HDF5 file for iterations >= iter_start
    and < iter_stop, optionally using another data manager instance instead of the
    global one returned by ``westpa.rc.get_data_manager()``.

    Returns a dictionary mapping target names (if available, target index otherwise)
    to a 1-D array of type ``fluxentry_dtype``, which contains columns for iteration
    number, flux, and count.
    '''

    data_manager = data_manager or westpa.rc.get_data_manager()
    iter_start = iter_start or 1
    iter_stop = iter_stop or data_manager.current_iteration

    if data_manager.we_h5file_version < 7:
        return _extract_fluxes_fileversion_lt_7(iter_start, iter_stop, data_manager)
    else:
        return _extract_fluxes_fileversion_7(iter_start, iter_stop, data_manager)


class WFluxanlTool(WESTTool):
    prog = 'w_fluxanl'
    description = '''\
Extract fluxes into pre-defined target states from WEST data,
average, and construct confidence intervals. Monte Carlo bootstrapping
is used to account for the correlated and possibly non-Gaussian statistical
error in flux measurements.

All non-graphical output (including that to the terminal and HDF5) assumes that
the propagation/resampling period ``tau`` is equal to unity; to obtain results
in familiar units, divide all fluxes and multiply all correlation lengths by
the true value of ``tau``.
'''

    output_format_version = 2

    def __init__(self):
        super().__init__()
        self.data_reader = WESTDataReader()
        self.iter_range = IterRangeSelection()
        self.output_h5file = None
        self.output_group = None
        self.target_groups = {}

        self.fluxdata = {}

        self.alpha = None
        self.autocorrel_alpha = None
        self.n_sets = None
        self.do_evol = False
        self.evol_step = 1

    def add_args(self, parser):
        self.data_reader.add_args(parser)
        self.iter_range.add_args(parser)
        ogroup = parser.add_argument_group('output options')
        ogroup.add_argument(
            '-o',
            '--output',
            default='fluxanl.h5',
            help='Store intermediate data and analysis results to OUTPUT (default: %(default)s).',
        )
        cgroup = parser.add_argument_group('calculation options')
        cgroup.add_argument(
            '--disable-bootstrap',
            '-db',
            dest='bootstrap',
            action='store_const',
            const=False,
            help='''Enable the use of Monte Carlo Block Bootstrapping.''',
        )
        cgroup.add_argument(
            '--disable-correl',
            '-dc',
            dest='correl',
            action='store_const',
            const=False,
            help='''Disable the correlation analysis.''',
        )
        cgroup.add_argument(
            '-a',
            '--alpha',
            type=float,
            default=0.05,
            help='''Calculate a (1-ALPHA) confidence interval on the average flux'
                             (default: %(default)s)''',
        )
        cgroup.add_argument(
            '--autocorrel-alpha',
            type=float,
            dest='acalpha',
            metavar='ACALPHA',
            help='''Evaluate autocorrelation of flux to (1-ACALPHA) significance.
                             Note that too small an ACALPHA will result in failure to detect autocorrelation
                             in a noisy flux signal. (Default: same as ALPHA.)''',
        )
        cgroup.add_argument(
            '-N', '--nsets', type=int, help='''Use NSETS samples for bootstrapping (default: chosen based on ALPHA)'''
        )
        cgroup.add_argument(
            '--evol',
            action='store_true',
            dest='do_evol',
            help='''Calculate time evolution of flux confidence intervals (expensive).''',
        )
        cgroup.add_argument(
            '--evol-step',
            type=int,
            default=1,
            metavar='ESTEP',
            help='''Calculate time evolution of flux confidence intervals every ESTEP
                            iterations (default: %(default)s)''',
        )

    def process_args(self, args):
        self.data_reader.process_args(args)
        self.data_reader.open()
        self.iter_range.data_manager = self.data_reader
        self.iter_range.process_args(args)

        self.output_h5file = h5py.File(args.output, 'w')

        self.alpha = args.alpha
        # Disable the bootstrap or the correlation analysis.
        self.mcbs_enable = args.bootstrap if args.bootstrap is not None else True
        self.do_correl = args.correl if args.correl is not None else True
        self.autocorrel_alpha = args.acalpha or self.alpha
        self.n_sets = args.nsets or mclib.get_bssize(self.alpha)

        self.do_evol = args.do_evol
        self.evol_step = args.evol_step or 1

    def calc_store_flux_data(self):
        westpa.rc.pstatus(
            'Calculating mean flux and confidence intervals for iterations [{},{})'.format(
                self.iter_range.iter_start, self.iter_range.iter_stop
            )
        )

        fluxdata = extract_fluxes(self.iter_range.iter_start, self.iter_range.iter_stop, self.data_reader)

        # Create a group to store data in
        output_group = h5io.create_hdf5_group(self.output_h5file, 'target_flux', replace=False, creating_program=self.prog)
        self.output_group = output_group
        output_group.attrs['version_code'] = self.output_format_version
        self.iter_range.record_data_iter_range(output_group)

        n_targets = len(fluxdata)
        index = np.empty((len(fluxdata),), dtype=target_index_dtype)
        avg_fluxdata = np.empty((n_targets,), dtype=ci_dtype)

        for itarget, (target_label, target_fluxdata) in enumerate(fluxdata.items()):
            # Create group and index entry
            index[itarget]['target_label'] = str(target_label)
            target_group = output_group.create_group('target_{}'.format(itarget))

            self.target_groups[target_label] = target_group

            # Store per-iteration values
            target_group['n_iter'] = target_fluxdata['n_iter']
            target_group['count'] = target_fluxdata['count']
            target_group['flux'] = target_fluxdata['flux']
            h5io.label_axes(target_group['flux'], ['n_iter'], units=['tau^-1'])

            # Calculate flux autocorrelation
            fluxes = target_fluxdata['flux']
            mean_flux = fluxes.mean()
            fmm = fluxes - mean_flux
            acorr = fftconvolve(fmm, fmm[::-1])
            acorr = acorr[len(acorr) // 2 :]
            acorr /= acorr[0]
            acorr_ds = target_group.create_dataset('flux_autocorrel', data=acorr)
            h5io.label_axes(acorr_ds, ['lag'], ['tau'])

            # Calculate overall averages and CIs
            # avg, lb_ci, ub_ci, correl_len = mclib.mcbs_ci_correl(fluxes, np.mean, self.alpha, self.n_sets,
            #                                                     autocorrel_alpha=self.autocorrel_alpha, subsample=np.mean)
            avg, lb_ci, ub_ci, sterr, correl_len = mclib.mcbs_ci_correl(
                {'dataset': fluxes},
                estimator=(lambda stride, dataset: np.mean(dataset)),
                alpha=self.alpha,
                n_sets=self.n_sets,
                autocorrel_alpha=self.autocorrel_alpha,
                subsample=np.mean,
                do_correl=self.do_correl,
                mcbs_enable=self.mcbs_enable,
            )
            avg_fluxdata[itarget] = (self.iter_range.iter_start, self.iter_range.iter_stop, avg, lb_ci, ub_ci, sterr, correl_len)
            westpa.rc.pstatus('target {!r}:'.format(target_label))
            westpa.rc.pstatus('  correlation length = {} tau'.format(correl_len))
            westpa.rc.pstatus('  mean flux and CI   = {:e} ({:e},{:e}) tau^(-1)'.format(avg, lb_ci, ub_ci))
            index[itarget]['mean_flux'] = avg
            index[itarget]['mean_flux_ci_lb'] = lb_ci
            index[itarget]['mean_flux_ci_ub'] = ub_ci
            index[itarget]['mean_flux_correl_len'] = correl_len

        # Write index and summary
        index_ds = output_group.create_dataset('index', data=index)
        index_ds.attrs['mcbs_alpha'] = self.alpha
        index_ds.attrs['mcbs_autocorrel_alpha'] = self.autocorrel_alpha
        index_ds.attrs['mcbs_n_sets'] = self.n_sets

        self.fluxdata = fluxdata
        self.output_h5file['avg_flux'] = avg_fluxdata

    def calc_evol_flux(self):
        westpa.rc.pstatus(
            'Calculating cumulative evolution of flux confidence intervals every {} iteration(s)'.format(self.evol_step)
        )

        for itarget, (target_label, target_fluxdata) in enumerate(self.fluxdata.items()):
            fluxes = target_fluxdata['flux']
            target_group = self.target_groups[target_label]
            iter_start = target_group['n_iter'][0]
            iter_stop = target_group['n_iter'][-1]
            iter_count = iter_stop - iter_start
            n_blocks = iter_count // self.evol_step
            if iter_count % self.evol_step > 0:
                n_blocks += 1

            cis = np.empty((n_blocks,), dtype=ci_dtype)

            for iblock in range(n_blocks):
                block_iter_stop = min(iter_start + (iblock + 1) * self.evol_step, iter_stop)
                istop = min((iblock + 1) * self.evol_step, len(target_fluxdata['flux']))
                fluxes = target_fluxdata['flux'][:istop]

                # avg, ci_lb, ci_ub, correl_len = mclib.mcbs_ci_correl(fluxes, np.mean, self.alpha, self.n_sets,
                #                                                     autocorrel_alpha = self.autocorrel_alpha,
                #                                                     subsample=np.mean)
                avg, ci_lb, ci_ub, sterr, correl_len = mclib.mcbs_ci_correl(
                    {'dataset': fluxes},
                    estimator=(lambda stride, dataset: np.mean(dataset)),
                    alpha=self.alpha,
                    n_sets=self.n_sets,
                    autocorrel_alpha=self.autocorrel_alpha,
                    subsample=np.mean,
                    do_correl=self.do_correl,
                    mcbs_enable=self.mcbs_enable,
                )
                cis[iblock]['iter_start'] = iter_start
                cis[iblock]['iter_stop'] = block_iter_stop
                cis[iblock]['expected'], cis[iblock]['ci_lbound'], cis[iblock]['ci_ubound'] = avg, ci_lb, ci_ub
                cis[iblock]['corr_len'] = correl_len
                cis[iblock]['sterr'] = sterr

                del fluxes

            cis_ds = target_group.create_dataset('flux_evolution', data=cis)
            cis_ds.attrs['iter_step'] = self.evol_step
            cis_ds.attrs['mcbs_alpha'] = self.alpha
            cis_ds.attrs['mcbs_autocorrel_alpha'] = self.autocorrel_alpha
            cis_ds.attrs['mcbs_n_sets'] = self.n_sets

    def go(self):
        self.calc_store_flux_data()
        if self.do_evol:
            self.calc_evol_flux()


def entry_point():
    warn('w_fluxanl is being deprecated.  Please use w_assign and w_direct instead.')
    WFluxanlTool().main()


if __name__ == '__main__':
    entry_point()



# import warnings
# warnings.filterwarnings('ignore', category=DeprecationWarning)
# warnings.filterwarnings('ignore', category=RuntimeWarning)
# warnings.filterwarnings('ignore', category=FutureWarning)
import numpy as np

# We don't care if we're reimporting blessings, under this context.
# warnings.filterwarnings('ignore')


class Plotter:
    '''
    This is a semi-generic plotting interface that has a built in curses based terminal plotter.
    It's fairly specific to what we're using it for here, but we could (and maybe should) build it out into
    a little library that we can use via the command line to plot things.  Might be useful for looking at data later.
    That would also cut the size of this tool down by a good bit.
    '''

    # def __init__(self, kinavg, kinrw, iteration, bin_labels, state_labels, state_pops, bin_pops, interface='matplotlib'):
    def __init__(self, h5file, h5key, iteration=-1, interface='matplotlib'):
        # Need to sort through and fix all this, but hey.
        self.iteration = iteration
        # These two are important for... reasons.
        try:
            # Commenting out below lines since `bin_labels` and `state_labels` are uninitialized variables.
            # This will always fail so just raise explicitly

            # self.bin_labels = list(bin_labels[...])
            # self.state_labels = list(state_labels[...]) + ['unknown']
            raise
        except Exception:
            try:
                self.state_labels = list(h5file['state_labels'][...]) + ['unknown']
            except Exception:
                self.state_labels = None
        # unless we totally fail out.
        self.interface = interface
        # What we should ACTUALLY do is just... yeah, just have it sub in what we need.
        # We'll need to throw in the state labels or whatever, but.
        self.h5file = h5file
        self.h5key = h5key
        # We should determine the number of dimensions of our dataset...
        # This has time data, so an i to j is a 3 dim, and an i is 2.
        try:
            self.dim = len(h5file[h5key].shape)
        except Exception:
            self.dim = 1
        try:
            # does the ci exist?
            _ = h5file[h5key]['expected']
        except Exception:
            self.dim = 1

    def plot(self, i=0, j=1, tau=1, iteration=None, dim=0, interface=None):
        if iteration is None:
            iteration = self.iteration
            self.__generic_ci__(self.h5file, iteration, i, j, tau=tau, h5key=self.h5key, dim=dim, interface=interface)

    def __generic_ci__(self, h5file, iteration, i, j, tau, h5key='rate_evolution', dim=0, interface=None):
        # This function just calls the appropriate plot function for our available
        # interface.
        if (interface is None and self.interface == 'text') or interface == 'text':
            if self.dim > 1:
                self.__terminal_ci__(h5file, iteration, i, j, tau, h5key)
            else:
                self.__terminal_expected__(h5file, iteration, i, j, tau, h5key, dim)
        else:
            try:
                import matplotlib

                matplotlib.use('TkAgg')
                from matplotlib import pyplot as plt

                if self.dim == 3:
                    plt.plot(h5file[h5key]['expected'][:iteration, i, j] / tau, color='black')
                    plt.plot(h5file[h5key]['ci_ubound'][:iteration, i, j] / tau, color='grey')
                    plt.plot(h5file[h5key]['ci_lbound'][:iteration, i, j] / tau, color='grey')
                else:
                    plt.plot(h5file[h5key]['expected'][:iteration, i] / tau, color='black')
                    plt.plot(h5file[h5key]['ci_ubound'][:iteration, i] / tau, color='grey')
                    plt.plot(h5file[h5key]['ci_lbound'][:iteration, i] / tau, color='grey')
                plt.show()
            except Exception:
                print('Unable to import plotting interface.  An X server ($DISPLAY) is required.')
                if self.dim > 1:
                    self.__terminal_ci__(h5file, iteration, i, j, tau)
                else:
                    self.__terminal_expected__(h5file, iteration, i, j, tau, h5key, dim)
                return 1

    def __generic_histo__(self, vector, labels):
        # This function just calls the appropriate plot function for our available
        # interface.  Same thing as generic_ci, but for a histogram.
        if self.interface == 'text':
            self.__terminal_histo__(vector, labels)
        else:
            try:
                import matplotlib

                matplotlib.use('TkAgg')
                from matplotlib import pyplot as plt

                plt.bar(
                    list(range(0, np.array(vector).shape[0])), vector, linewidth=0, align='center', color='gold', tick_label=labels
                )
                plt.show()
            except Exception:
                print('Unable to import plotting interface.  An X server ($DISPLAY) is required.')
                self.__terminal_histo__(vector, labels)
                return 1

    def __terminal_histo__(self, vector, labels, fullscreen_mode=True):
        from blessings import Terminal

        self.t = Terminal()
        h = int(self.t.height / 4) * 3
        w = self.t.width
        cols = np.array(vector).shape[0]
        # Let's print this business!

        colwidth = w / cols
        with self.t.fullscreen():
            for y in range(0, h):
                for x in range(0, cols):
                    if x == 0:
                        with self.t.location(0, y):
                            print(self.t.red('{0:.4f}|'.format(float(h - y) / float(h))))
                    with self.t.location((x * colwidth) + 8 + len(labels[x]) / 2, y):
                        if vector[x] >= (float(h - y) / float(h)):
                            # print(float(h-y)/float(h))
                            print(self.t.on_blue(' '))
            for x in range(0, cols):
                if x == 0:
                    with self.t.location(x, h):
                        print('States| ')
                with self.t.location((x * colwidth) + 8, h):
                    print(self.t.blue(labels[x]))

            if fullscreen_mode:
                input("Press enter to continue.")

    def __terminal_ci__(self, h5file, iteration, si, sj, tau, h5key):
        from blessings import Terminal

        self.t = Terminal()
        h = int(self.t.height / 4 * 3.75)
        # We'll figure out how to subsample the timepoints...
        w = self.t.width
        if self.dim == 3:
            in_tup = (iteration - 1, si, sj)
        else:
            in_tup = (iteration - 1, si)
        yupper = (h5file[h5key]['ci_ubound'][in_tup] / tau) * 2
        ylower = (h5file[h5key]['ci_lbound'][in_tup] / tau) / 2
        # Here are points pertaining to height.
        scale = np.array([0.0] + [ylower + i * (yupper - ylower) / np.float(h) for i in range(0, h)])[::-1]
        if iteration > w:
            block_size = iteration / w
        else:
            block_size = 1

        with self.t.fullscreen():
            try:
                for x in range(0, w - 12):
                    iter = x * block_size
                    if self.dim == 3:
                        in_tup = (iter - 1, si, sj)
                    else:
                        in_tup = (iter - 1, si)
                    yupper = h5file[h5key]['ci_ubound'][in_tup] / tau
                    ylower = h5file[h5key]['ci_lbound'][in_tup] / tau
                    ci = np.digitize([yupper, ylower], scale)
                    if x == 0:
                        for y in range(0, h + 1):
                            with self.t.location(0, y):
                                print(self.t.bold(self.t.red('{0:.7f}|'.format(scale[y]))))
                    for y in range(ci[0], ci[1]):
                        # with self.t.location(x+12, y):
                        print(self.t.move(y, x + 12) + self.t.on_blue(' '))
                        # print(self.t.on_blue(' '))
                    # with self.t.location(x+12, np.digitize(h5file['rate_evolution']['expected'][iter-1, si, sj]/tau, scale)):
                    #        print(self.t.on_blue('-'))
                    print(self.t.move(np.digitize(h5file[h5key]['expected'][in_tup] / tau, scale), x + 12) + self.t.on_blue('-'))

                for x in range(0, w - 12, w / 10):
                    if x == 0:
                        with self.t.location(x, h + 1):
                            print('Iteration| ')
                    with self.t.location(x + 12, h + 1):
                        iter = x * block_size
                        print(self.t.blue(str(iter)))
            except Exception:
                pass

            with self.t.location(0, h + 2):
                # We need to improve this.
                # if h5key == 'rate_evolution':
                #    print("k_ij from {} to {} from iter 1 to {}".format(self.state_labels[si], self.state_labels[sj], self.iteration))
                # elif h5key == 'conditional_flux_evolution':
                #    print("i->j flux from {} to {} from iter 1 to {}".format(self.state_labels[si], self.state_labels[sj], self.iteration))
                if self.dim == 3:
                    print(
                        "{} from {} to {} from iter 1 to {}".format(
                            h5key, self.state_labels[si], self.state_labels[sj], self.iteration
                        )
                    )
                else:
                    print("{} of state {} from iter 1 to {}".format(h5key, self.state_labels[si], self.iteration))
            with self.t.location(0, h + 3):
                input("Press enter to continue.")

    def __terminal_expected__(self, h5file, iteration, si, sj, tau, h5key, dim):
        from blessings import Terminal

        self.t = Terminal()
        h = int(self.t.height / 4 * 3.75)
        # We'll figure out how to subsample the timepoints...
        w = self.t.width
        if self.dim == 3:
            in_tup = (iteration - 1, si, sj)
        else:
            in_tup = (iteration - 1, si)
        in_tup = (iteration - 1, dim)
        try:
            yupper = (np.max(h5file) / tau) * 2
        except Exception:
            in_tup = iteration - 1
            yupper = (np.max(h5file) / tau) * 2
        ylower = (np.min(h5file) / tau) * 2
        # Here are points pertaining to height.
        if yupper > 0:
            yupper = (np.max(h5file) / tau) * 1.2
            ylower = (np.min(h5file) / tau) / 2
            scale = np.array([0.0] + [ylower + i * (yupper - ylower) / np.float(h) for i in range(0, h)])[::-1]
        else:
            yupper = (np.max(h5file) / tau) / 2
            ylower = (np.min(h5file) / tau) * 1.2
            scale = np.array([ylower + i * (yupper - ylower) / np.float(h) for i in range(0, h)] + [0.0])[::-1]
        if iteration > w:
            block_size = iteration / w
        else:
            block_size = 1

        with self.t.fullscreen():
            try:
                for x in range(0, w - 12):
                    iter = x * block_size
                    if self.dim == 3:
                        in_tup = (iter - 1, si, sj)
                    else:
                        in_tup = (iter - 1, si)
                    in_tup = (iter - 1, dim)
                    try:
                        yupper = h5file[in_tup] / tau
                    except Exception:
                        in_tup = iter - 1
                        yupper = h5file[in_tup] / tau
                    ylower = h5file[in_tup] / tau
                    ci = np.digitize([yupper, ylower], scale)
                    if x == 0:
                        for y in range(0, h + 1):
                            with self.t.location(0, y):
                                print(self.t.bold(self.t.red('{0:.7f}|'.format(scale[y]))))
                    for y in range(ci[0], ci[1]):
                        print(self.t.move(y, x + 12) + self.t.on_blue(' '))
                        # print(self.t.on_blue(' '))
                    print(self.t.move(np.digitize(h5file[in_tup] / tau, scale), x + 12) + self.t.on_blue('-'))

                for x in range(0, w - 12, w / 10):
                    if x == 0:
                        with self.t.location(x, h + 1):
                            print('Iteration| ')
                    with self.t.location(x + 12, h + 1):
                        iter = x * block_size
                        print(self.t.blue(str(iter)))
            except Exception:
                pass

            with self.t.location(0, h + 2):
                # We need to improve this.
                print("{} from iter 1 to {}".format(h5key, self.iteration))
            with self.t.location(0, h + 3):
                input("Press enter to continue.")



import logging

import h5py
import numpy as np

from westpa.tools import (
    WESTParallelTool,
    WESTDataReader,
    WESTDSSynthesizer,
    WESTWDSSynthesizer,
    IterRangeSelection,
    ProgressIndicatorComponent,
)

from westpa.fasthist import histnd, normhistnd
from westpa.core import h5io


log = logging.getLogger('w_pdist')


def isiterable(x):
    try:
        iter(x)
    except TypeError:
        return False
    else:
        return True


def _remote_min_max(ndim, dset_dtype, n_iter, dsspec):
    try:
        minval = np.finfo(dset_dtype).min
        maxval = np.finfo(dset_dtype).max
    except ValueError:
        minval = np.iinfo(dset_dtype).min
        maxval = np.iinfo(dset_dtype).max

    data_range = [(maxval, minval) for _i in range(ndim)]

    dset = dsspec.get_iter_data(n_iter)
    for idim in range(ndim):
        dimdata = dset[:, :, idim]
        current_min, current_max = data_range[idim]
        current_min = min(current_min, dimdata.min())
        current_max = max(current_max, dimdata.max())
        data_range[idim] = (current_min, current_max)
        del dimdata
    del dset
    return data_range


def _remote_bin_iter(iiter, n_iter, dsspec, wt_dsspec, initpoint, binbounds, ignore_out_of_range):

    iter_hist_shape = tuple(len(bounds) - 1 for bounds in binbounds)
    iter_hist = np.zeros(iter_hist_shape, dtype=np.float64)

    dset = dsspec.get_iter_data(n_iter)
    npts = dset.shape[1]
    weights = wt_dsspec.get_iter_data(n_iter)

    dset = dset[:, initpoint:, :]
    for ipt in range(npts - initpoint):
        histnd(dset[:, ipt, :], binbounds, weights, out=iter_hist, binbound_check=False, ignore_out_of_range=ignore_out_of_range)

    del weights, dset

    # normalize histogram
    normhistnd(iter_hist, binbounds)
    return iiter, n_iter, iter_hist


class WPDist(WESTParallelTool):
    prog = 'w_pdist'
    description = '''\
Calculate time-resolved, multi-dimensional probability distributions of WE
datasets.


-----------------------------------------------------------------------------
Source data
-----------------------------------------------------------------------------

Source data is provided either by a user-specified function
(--construct-dataset) or a list of "data set specifications" (--dsspecs).
If neither is provided, the progress coordinate dataset ''pcoord'' is used.

To use a custom function to extract or calculate data whose probability
distribution will be calculated, specify the function in standard Python
MODULE.FUNCTION syntax as the argument to --construct-dataset. This function
will be called as function(n_iter,iter_group), where n_iter is the iteration
whose data are being considered and iter_group is the corresponding group
in the main WEST HDF5 file (west.h5). The function must return data which can
be indexed as [segment][timepoint][dimension].

To use a list of data set specifications, specify --dsspecs and then list the
desired datasets one-by-one (space-separated in most shells). These data set
specifications are formatted as NAME[,file=FILENAME,slice=SLICE], which will
use the dataset called NAME in the HDF5 file FILENAME (defaulting to the main
WEST HDF5 file west.h5), and slice it with the Python slice expression SLICE
(as in [0:2] to select the first two elements of the first axis of the
dataset). The ``slice`` option is most useful for selecting one column (or
more) from a multi-column dataset, such as arises when using a progress
coordinate of multiple dimensions.


-----------------------------------------------------------------------------
Histogram binning
-----------------------------------------------------------------------------

By default, histograms are constructed with 100 bins in each dimension. This
can be overridden by specifying -b/--bins, which accepts a number of different
kinds of arguments:

  a single integer N
    N uniformly spaced bins will be used in each dimension.

  a sequence of integers N1,N2,... (comma-separated)
    N1 uniformly spaced bins will be used for the first dimension, N2 for the
    second, and so on.

  a list of lists [[B11, B12, B13, ...], [B21, B22, B23, ...], ...]
    The bin boundaries B11, B12, B13, ... will be used for the first dimension,
    B21, B22, B23, ... for the second dimension, and so on. These bin
    boundaries need not be uniformly spaced. These expressions will be
    evaluated with Python's ``eval`` construct, with ``np`` available for
    use [e.g. to specify bins using np.arange()].

The first two forms (integer, list of integers) will trigger a scan of all
data in each dimension in order to determine the minimum and maximum values,
which may be very expensive for large datasets. This can be avoided by
explicitly providing bin boundaries using the list-of-lists form.

Note that these bins are *NOT* at all related to the bins used to drive WE
sampling.


-----------------------------------------------------------------------------
Output format
-----------------------------------------------------------------------------

The output file produced (specified by -o/--output, defaulting to "pdist.h5")
may be fed to plothist to generate plots (or appropriately processed text or
HDF5 files) from this data. In short, the following datasets are created:

  ``histograms``
    Normalized histograms. The first axis corresponds to iteration, and
    remaining axes correspond to dimensions of the input dataset.

  ``/binbounds_0``
    Vector of bin boundaries for the first (index 0) dimension. Additional
    datasets similarly named (/binbounds_1, /binbounds_2, ...) are created
    for additional dimensions.

  ``/midpoints_0``
    Vector of bin midpoints for the first (index 0) dimension. Additional
    datasets similarly named are created for additional dimensions.

  ``n_iter``
    Vector of iteration numbers corresponding to the stored histograms (i.e.
    the first axis of the ``histograms`` dataset).


-----------------------------------------------------------------------------
Subsequent processing
-----------------------------------------------------------------------------

The output generated by this program (-o/--output, default "pdist.h5") may be
plotted by the ``plothist`` program. See ``plothist --help`` for more
information.


-----------------------------------------------------------------------------
Parallelization
-----------------------------------------------------------------------------

This tool supports parallelized binning, including reading of input data.
Parallel processing is the default. For simple cases (reading pre-computed
input data, modest numbers of segments), serial processing (--serial) may be
more efficient.


-----------------------------------------------------------------------------
Command-line options
-----------------------------------------------------------------------------

'''

    def __init__(self):
        super().__init__()

        # Parallel processing by default (this is not actually necessary, but it is
        # informative!)
        self.wm_env.default_work_manager = self.wm_env.default_parallel_work_manager

        # These are used throughout
        self.progress = ProgressIndicatorComponent()
        self.data_reader = WESTDataReader()
        self.input_dssynth = WESTDSSynthesizer(default_dsname='pcoord')
        self.input_wdssynth = WESTWDSSynthesizer(default_dsname='seg_index')
        self.iter_range = IterRangeSelection(self.data_reader)
        self.iter_range.include_args['iter_step'] = False
        self.binspec = None
        self.output_filename = None
        self.output_file = None

        self.dsspec = None
        self.wt_dsspec = None  # dsspec for weights

        # These are used during histogram generation only
        self.iter_start = None
        self.iter_stop = None
        self.ndim = None
        self.ntimepoints = None
        self.dset_dtype = None
        self.binbounds = None  # bin boundaries for each dimension
        self.midpoints = None  # bin midpoints for each dimension
        self.data_range = None  # data range for each dimension, as the pairs (min,max)
        self.ignore_out_of_range = False
        self.compress_output = False

    def add_args(self, parser):
        self.data_reader.add_args(parser)

        self.iter_range.add_args(parser)

        parser.add_argument(
            '-b',
            '--bins',
            dest='bins',
            metavar='BINEXPR',
            default='100',
            help='''Use BINEXPR for bins. This may be an integer, which will be used for each
                            dimension of the progress coordinate; a list of integers (formatted as [n1,n2,...])
                            which will use n1 bins for the first dimension, n2 for the second dimension, and so on;
                            or a list of lists of boundaries (formatted as [[a1, a2, ...], [b1, b2, ...], ... ]), which
                            will use [a1, a2, ...] as bin boundaries for the first dimension, [b1, b2, ...] as bin boundaries
                            for the second dimension, and so on. (Default: 100 bins in each dimension.)''',
        )

        parser.add_argument(
            '-o', '--output', dest='output', default='pdist.h5', help='''Store results in OUTPUT (default: %(default)s).'''
        )
        parser.add_argument(
            '-C',
            '--compress',
            action='store_true',
            help='''Compress histograms. May make storage of higher-dimensional histograms
                            more tractable, at the (possible extreme) expense of increased analysis time.
                            (Default: no compression.)''',
        )

        parser.add_argument(
            '--loose',
            dest='ignore_out_of_range',
            action='store_true',
            help='''Ignore values that do not fall within bins. (Risky, as this can make buggy bin
                            boundaries appear as reasonable data. Only use if you are
                            sure of your bin boundary specification.)''',
        )

        igroup = parser.add_argument_group('input dataset options').add_mutually_exclusive_group(required=False)

        igroup.add_argument(
            '--construct-dataset',
            help='''Use the given function (as in module.function) to extract source data.
                            This function will be called once per iteration as function(n_iter, iter_group)
                            to construct data for one iteration. Data returned must be indexable as
                            [seg_id][timepoint][dimension]''',
        )

        igroup.add_argument(
            '--dsspecs', nargs='+', metavar='DSSPEC', help='''Construct probability distribution from one or more DSSPECs.'''
        )

        wgroup = parser.add_argument_group('input weight dataset options').add_mutually_exclusive_group(required=False)
        wgroup.add_argument(
            '--construct-wdataset',
            help='''Use the given function (as in module.function) to extract weight data.
                            This function will be called once per iteration as function(n_iter, iter_group)
                            to construct data for one iteration. Data returned must be indexable as
                            [seg_id]''',
        )

        self.progress.add_args(parser)

    def process_args(self, args):
        self.progress.process_args(args)
        self.data_reader.process_args(args)
        self.input_dssynth.h5filename = self.data_reader.we_h5filename
        self.input_dssynth.process_args(args)
        self.dsspec = self.input_dssynth.dsspec

        # Carrying an open HDF5 file across a fork() seems to corrupt the entire HDF5 library
        # Open the WEST HDF5 file just long enough to process our iteration range, then close
        # and reopen in go() [which executes after the fork]
        with self.data_reader:
            self.iter_range.process_args(args)

        # Reading potential custom weights
        self.input_wdssynth.h5filename = self.data_reader.we_h5filename
        self.input_wdssynth.process_args(args)
        self.wt_dsspec = self.input_wdssynth.dsspec

        self.binspec = args.bins
        self.output_filename = args.output
        self.ignore_out_of_range = bool(args.ignore_out_of_range)
        self.compress_output = args.compress or False

    def go(self):
        self.data_reader.open('r')
        pi = self.progress.indicator
        pi.operation = 'Initializing'
        with pi:
            self.output_file = h5py.File(self.output_filename, 'w')
            h5io.stamp_creator_data(self.output_file)

            self.iter_start = self.iter_range.iter_start
            self.iter_stop = self.iter_range.iter_stop

            # Construct bin boundaries
            self.construct_bins(self.parse_binspec(self.binspec))
            for idim, (binbounds, midpoints) in enumerate(zip(self.binbounds, self.midpoints)):
                self.output_file['binbounds_{}'.format(idim)] = binbounds
                self.output_file['midpoints_{}'.format(idim)] = midpoints

            # construct histogram
            self.construct_histogram()

            # Record iteration range
            iter_range = self.iter_range.iter_range()
            self.output_file['n_iter'] = iter_range
            self.iter_range.record_data_iter_range(self.output_file['histograms'])

            self.output_file.close()

    @staticmethod
    def parse_binspec(binspec):
        namespace = {'numpy': np, 'np': np, 'inf': float('inf')}

        try:
            binspec_compiled = eval(binspec, namespace)
        except Exception as e:
            raise ValueError('invalid bin specification: {!r}'.format(e))
        else:
            if log.isEnabledFor(logging.DEBUG):
                log.debug('bin specs: {!r}'.format(binspec_compiled))
        return binspec_compiled

    def construct_bins(self, bins):
        '''
        Construct bins according to ``bins``, which may be:

          1) A scalar integer (for that number of bins in each dimension)
          2) A sequence of integers (specifying number of bins for each dimension)
          3) A sequence of sequences of bin boundaries (specifying boundaries for each dimension)

        Sets ``self.binbounds`` to a list of arrays of bin boundaries appropriate for passing to
        fasthist.histnd, along with ``self.midpoints`` to the midpoints of the bins.
        '''

        if not isiterable(bins):
            self._construct_bins_from_scalar(bins)
        elif not isiterable(bins[0]):
            self._construct_bins_from_int_seq(bins)
        else:
            self._construct_bins_from_bound_seqs(bins)

        if log.isEnabledFor(logging.DEBUG):
            log.debug('binbounds: {!r}'.format(self.binbounds))

    def scan_data_shape(self):
        if self.ndim is None:
            dset = self.dsspec.get_iter_data(self.iter_start)
            self.ntimepoints = dset.shape[1]
            self.ndim = dset.shape[2]
            self.dset_dtype = dset.dtype

    def scan_data_range(self):
        '''Scan input data for range in each dimension. The number of dimensions is determined
        from the shape of the progress coordinate as of self.iter_start.'''

        self.progress.indicator.new_operation('Scanning for data range', self.iter_stop - self.iter_start)
        self.scan_data_shape()

        dset_dtype = self.dset_dtype
        ndim = self.ndim
        dsspec = self.dsspec

        try:
            minval = np.finfo(dset_dtype).min
            maxval = np.finfo(dset_dtype).max
        except ValueError:
            minval = np.iinfo(dset_dtype).min
            maxval = np.iinfo(dset_dtype).max

        data_range = self.data_range = [(maxval, minval) for _i in range(self.ndim)]

        # futures = []
        # for n_iter in xrange(self.iter_start, self.iter_stop):
        # _remote_min_max(ndim, dset_dtype, n_iter, dsspec)
        #    futures.append(self.work_manager.submit(_remote_min_max, args=(ndim, dset_dtype, n_iter, dsspec)))

        # for future in self.work_manager.as_completed(futures):
        for future in self.work_manager.submit_as_completed(
            ((_remote_min_max, (ndim, dset_dtype, n_iter, dsspec), {}) for n_iter in range(self.iter_start, self.iter_stop)),
            self.max_queue_len,
        ):
            bounds = future.get_result(discard=True)
            for idim in range(ndim):
                current_min, current_max = data_range[idim]
                current_min = min(current_min, bounds[idim][0])
                current_max = max(current_max, bounds[idim][1])
                data_range[idim] = (current_min, current_max)
            self.progress.indicator.progress += 1

    def _construct_bins_from_scalar(self, bins):
        if self.data_range is None:
            self.scan_data_range()

        self.binbounds = []
        self.midpoints = []
        for idim in range(self.ndim):
            lb, ub = self.data_range[idim]
            # Advance just beyond the upper bound of the range, so that we catch
            # the maximum in the histogram
            ub *= 1.01

            boundset = np.linspace(lb, ub, bins + 1)
            midpoints = (boundset[:-1] + boundset[1:]) / 2.0
            self.binbounds.append(boundset)
            self.midpoints.append(midpoints)

    def _construct_bins_from_int_seq(self, bins):
        if self.data_range is None:
            self.scan_data_range()

        self.binbounds = []
        self.midpoints = []
        for idim in range(self.ndim):
            lb, ub = self.data_range[idim]
            # Advance just beyond the upper bound of the range, so that we catch
            # the maximum in the histogram
            ub *= 1.01

            boundset = np.linspace(lb, ub, bins[idim] + 1)
            midpoints = (boundset[:-1] + boundset[1:]) / 2.0
            self.binbounds.append(boundset)
            self.midpoints.append(midpoints)

    def _construct_bins_from_bound_seqs(self, bins):
        self.binbounds = []
        self.midpoints = []
        for boundset in bins:
            boundset = np.asarray(boundset)
            if (np.diff(boundset) <= 0).any():
                raise ValueError('boundary set {!r} is not strictly monotonically increasing'.format(boundset))
            self.binbounds.append(boundset)
            self.midpoints.append((boundset[:-1] + boundset[1:]) / 2.0)

    def construct_histogram(self):
        '''Construct a histogram using bins previously constructed with ``construct_bins()``.
        The time series of histogram values is stored in ``histograms``.
        Each histogram in the time series is normalized.'''

        self.scan_data_shape()

        iter_count = self.iter_stop - self.iter_start
        histograms_ds = self.output_file.create_dataset(
            'histograms',
            dtype=np.float64,
            shape=((iter_count,) + tuple(len(bounds) - 1 for bounds in self.binbounds)),
            compression=9 if self.compress_output else None,
        )
        binbounds = [np.require(boundset, self.dset_dtype, 'C') for boundset in self.binbounds]

        self.progress.indicator.new_operation('Constructing histograms', self.iter_stop - self.iter_start)
        task_gen = (
            (
                _remote_bin_iter,
                (iiter, n_iter, self.dsspec, self.wt_dsspec, 1 if iiter > 0 else 0, binbounds, self.ignore_out_of_range),
                {},
            )
            for (iiter, n_iter) in enumerate(range(self.iter_start, self.iter_stop))
        )
        # futures = set()
        # for iiter, n_iter in enumerate(xrange(self.iter_start, self.iter_stop)):
        #    initpoint = 1 if iiter > 0 else 0
        #    futures.add(self.work_manager.submit(_remote_bin_iter,
        #                                            args=(iiter, n_iter, self.dsspec, self.wt_dsspec, initpoint, binbounds)))

        # for future in self.work_manager.as_completed(futures):
        # future = self.work_manager.wait_any(futures)
        # for future in self.work_manager.submit_as_completed(task_gen, self.queue_size):
        log.debug('max queue length: {!r}'.format(self.max_queue_len))
        for future in self.work_manager.submit_as_completed(task_gen, self.max_queue_len):
            iiter, n_iter, iter_hist = future.get_result(discard=True)
            self.progress.indicator.progress += 1

            # store histogram
            histograms_ds[iiter] = iter_hist
            del iter_hist, future


def entry_point():
    WPDist().main()


if __name__ == '__main__':
    entry_point()



'''Core classes for creating WESTPA command-line tools'''

import argparse
import logging
import sys
import os

import westpa
from westpa import work_managers
from westpa.core import h5io


log = logging.getLogger(__name__)


class WESTToolComponent:
    '''Base class for WEST command line tools and components used in constructing tools'''

    def __init__(self):
        self.config_required = False
        self.include_args = {}
        self.arg_defaults = {}
        self.parser = None
        self.args = None

    def include_arg(self, argname):
        self.include_args[argname] = True

    def exclude_arg(self, argname):
        self.include_args[argname] = False

    def set_arg_default(self, argname, value):
        self.arg_defaults[argname] = value

    def add_args(self, parser):
        '''Add arguments specific to this component to the given argparse parser.'''
        pass

    def process_args(self, args):
        '''Take argparse-processed arguments associated with this component and deal
        with them appropriately (setting instance variables, etc)'''
        pass

    def add_all_args(self, parser):
        '''Add arguments for all components from which this class derives to the given parser,
        starting with the class highest up the inheritance chain (most distant ancestor).'''
        self.parser = parser
        for cls in reversed(self.__class__.__mro__):
            try:
                fn = cls.__dict__['add_args']
            except KeyError:
                pass
            else:
                fn(self, parser)

    def process_all_args(self, args):
        self.args = args
        '''Process arguments for all components from which this class derives,
        starting with the class highest up the inheritance chain (most distant ancestor).'''
        for cls in reversed(self.__class__.__mro__):
            try:
                fn = cls.__dict__['process_args']
            except KeyError:
                pass
            else:
                fn(self, args)


class WESTTool(WESTToolComponent):
    '''Base class for WEST command line tools'''

    prog = None
    usage = None
    description = None
    epilog = None

    def __init__(self):
        super().__init__()

    def add_args(self, parser):
        '''Add arguments specific to this tool to the given argparse parser.'''
        westpa.rc.add_args(parser)

    def process_args(self, args):
        '''Take argparse-processed arguments associated with this tool and deal
        with them appropriately (setting instance variables, etc)'''
        westpa.rc.process_args(args, config_required=self.config_required)

    def make_parser(self, prog=None, usage=None, description=None, epilog=None, args=None):
        prog = prog or self.prog
        usage = usage or self.usage
        description = description or self.description
        epilog = epilog or self.epilog
        parser = argparse.ArgumentParser(
            prog=prog,
            usage=usage,
            description=description,
            epilog=epilog,
            formatter_class=argparse.RawDescriptionHelpFormatter,
            conflict_handler='resolve',
        )
        self.add_all_args(parser)
        return parser

    def make_parser_and_process(self, prog=None, usage=None, description=None, epilog=None, args=None):
        '''A convenience function to create a parser, call add_all_args(), and then call process_all_args().
        The argument namespace is returned.'''
        parser = self.make_parser(prog, usage, description, epilog, args)
        args = parser.parse_args(args)
        self.process_all_args(args)
        return args

    def go(self):
        '''Perform the analysis associated with this tool.'''
        raise NotImplementedError

    def main(self):
        '''A convenience function to make a parser, parse and process arguments, then call self.go()'''
        self.make_parser_and_process()
        self.go()


class WESTParallelTool(WESTTool):
    '''Base class for command-line tools parallelized with wwmgr. This automatically adds and processes
    wwmgr command-line arguments and creates a work manager at self.work_manager.'''

    def __init__(self, wm_env=None):
        super().__init__()
        self.work_manager = None
        self.wm_env = wm_env or work_managers.environment.default_env
        self.max_queue_len = None

    def make_parser_and_process(self, prog=None, usage=None, description=None, epilog=None, args=None):
        '''A convenience function to create a parser, call add_all_args(), and then call process_all_args().
        The argument namespace is returned.'''
        parser = self.make_parser(prog, usage, description, epilog, args)
        self.wm_env.add_wm_args(parser)

        args = parser.parse_args(args)
        self.wm_env.process_wm_args(args)

        # Instantiate work manager
        self.work_manager = self.wm_env.make_work_manager()

        # Process args
        self.process_all_args(args)
        return args

    def add_args(self, parser):
        pgroup = parser.add_argument_group('parallelization options')
        pgroup.add_argument(
            '--max-queue-length',
            type=int,
            help='''Maximum number of tasks that can be queued. Useful to limit RAM use
                            for tasks that have very large requests/response. Default: no limit.''',
        )

    def process_args(self, args):
        self.max_queue_len = args.max_queue_length
        log.debug('max queue length: {!r}'.format(self.max_queue_len))

    def go(self):
        '''Perform the analysis associated with this tool.'''
        raise NotImplementedError

    def main(self):
        '''A convenience function to make a parser, parse and process arguments, then run self.go() in the master process.'''
        self.make_parser_and_process()
        with self.work_manager:
            if self.work_manager.is_master:
                self.go()
            else:
                self.work_manager.run()


class WESTMultiTool(WESTParallelTool):
    '''Base class for command-line tools which work with multiple simulations.  Automatically parses for
    and gives commands to load multiple files.'''

    def __init__(self, wm_env=None):
        super(WESTTool, self).__init__()
        self.ntrials = 0
        self.master = None

    def make_parser_and_process(self, prog=None, usage=None, description=None, epilog=None, args=None):
        '''A convenience function to create a parser, call add_all_args(), and then call process_all_args().
        The argument namespace is returned.'''
        parser = self.make_parser(prog, usage, description, epilog, args)
        args = parser.parse_args(args)

        # Process args
        self.process_all_args(args)
        return args

    def parse_from_yaml(self, yamlfilepath):
        '''Parse options from YAML input file. Command line arguments take
        precedence over options specified in the YAML hierarchy.
        TODO: add description on how YAML files should be constructed.
        '''
        import yaml

        with open(yamlfilepath, 'r') as yamlfile:
            self.yamlargdict = yaml.load(yamlfile, Loader=yaml.Loader)
        # add in options here for intelligent processing of files

    def add_args(self, parser):
        mgroup = parser.add_argument_group('multiple simulation options')
        mgroup.add_argument(
            '-m',
            '--master',
            default=os.getcwd(),
            help='''Master path of simulations; this is where all the small sims
                             are stored.''',
        )
        mgroup.add_argument('-n', '--sims', default=0, help='''The number of simulation directories.  Assumes leading zeros.''')

    class NoSimulationsException(Exception):
        pass

    def generate_file_list(self, key_list):
        '''A convenience function which takes in a list of keys that are filenames, and returns a dictionary
        which contains all the individual files loaded inside of a dictionary keyed to the filename.'''
        return_dict = {}
        if self.ntrials == 0:
            raise self.NoSimulationsException('You must specify the number of simulations.')

        for key in key_list:
            return_dict[key] = {}
        for i in range(1, self.ntrials + 1):
            # Need to not make this hard coded, but who cares for now.
            for key in key_list:
                return_dict[key][i] = h5io.WESTPAH5File(os.path.join(self.master, str(i).zfill(2), key), 'r')
        return return_dict

    def process_args(self, args):
        self.master = args.master
        self.ntrials = int(args.sims)
        log.debug('Simulations loaded from: {!r}'.format(self.master))
        log.debug('Number of simulations: {!r}'.format(self.ntrials))

    def go(self):
        '''Perform the analysis associated with this tool.'''
        raise NotImplementedError

    def main(self):
        '''A convenience function to make a parser, parse and process arguments, then run self.go() in the master process.'''
        self.make_parser_and_process()
        self.go()


class WESTSubcommand(WESTToolComponent):
    '''Base class for command-line tool subcommands. A little sugar for making this
    more uniform.'''

    subcommand = None
    help_text = None
    description = None

    def __init__(self, parent):
        self.parent = parent
        self.subparser = None

    def add_to_subparsers(self, subparsers):
        subparser = subparsers.add_parser(
            self.subcommand,
            help=self.help_text,
            description=self.description,
            formatter_class=argparse.RawDescriptionHelpFormatter,
        )
        self.add_all_args(subparser)
        subparser.set_defaults(west_subcommand=self)
        self.subparser = subparser

    def go(self):
        raise NotImplementedError

    @property
    def work_manager(self):
        '''The work manager for this tool. Raises AttributeError if this is not a parallel
        tool.'''
        return self.parent.work_manager


class _WESTSubcommandHelp(WESTSubcommand):
    subcommand = 'help'
    help_text = 'print help for this command or individual subcommands'

    def add_args(self, parser):
        parser.add_argument('command', nargs='?', choices=[subcommand.subcommand for subcommand in self.parent.subcommands])

    def process_args(self, args):
        self.command = args.command

    def go(self):
        if self.command is None:
            # Get parent help
            self.parent.parser.print_help(sys.stdout)
        else:
            self.parent._subcommand_instances[self.command].subparser.print_help(sys.stdout)
        sys.exit(0)


class WESTMasterCommand(WESTTool):
    '''Base class for command-line tools that employ subcommands'''

    subparsers_title = None
    subcommands = None

    include_help_command = True

    def __init__(self):
        super().__init__()
        self._subcommand = None
        self._subcommand_instances = {subcommand_class.subcommand: subcommand_class(self) for subcommand_class in self.subcommands}

        # Sanity checks
        if __debug__:
            for scclass in self.subcommands:
                assert scclass.subcommand, 'subcommand {!r} does not define a "subcommand" class variable'.format(scclass)

    def add_args(self, parser):
        subparsers = parser.add_subparsers(title=self.subparsers_title)
        if self.include_help_command:
            _WESTSubcommandHelp(self).add_to_subparsers(subparsers)
        for instance in self._subcommand_instances.values():
            instance.add_to_subparsers(subparsers)

    def process_args(self, args):
        try:
            self._subcommand = args.west_subcommand
        except AttributeError:
            # No subcommand given; display help
            print('Error: a command is required. See below.', file=sys.stderr)
            self.parser.print_help(sys.stderr)
            sys.exit(2)
        else:
            self._subcommand.process_all_args(args)

    def go(self):
        self._subcommand.go()



import numpy as np
import scipy
import scipy.optimize


class FourierFit:
    def __init__(self, P=2, ndims=2, maxiters=100, tol=1.0e-6):
        super().__init__()

        self.P = P
        self.maxiters = maxiters
        self.ndims = ndims
        self.tol = tol
        self.pp = []
        self.t0 = None
        self.w0 = None

    def calc_string(self, w, t, x_meas):
        tlen = len(t)
        t = np.linspace(0.0, 1.0, tlen)
        x_est = x_meas[0, :] + (x_meas[-1, :] - x_meas[0, :]) * t[:, np.newaxis]
        for i in range(self.ndims):
            for j in range(self.P):
                x_est[:, i] += w[i, j] * np.sin((j + 1) * np.pi * t)
        return x_est

    def _optimize_dist(self, tk, x_meas, w, k):
        x_target = x_meas[k, :]
        x_est = x_meas[0, :] + (x_meas[-1, :] - x_meas[0, :]) * tk
        for i in range(self.ndims):
            for j in range(self.P):
                x_est[i] += w[i, j] * np.sin((j + 1) * np.pi * tk)

        err = x_target - x_est
        return err

    def _optimize_w(self, w, x_meas, t, k, weight):
        x_target = x_meas[:, k]
        x_est = x_meas[0, k] + (x_meas[-1, k] - x_meas[0, k]) * t

        for j in range(self.P):
            x_est += w[j] * np.sin((j + 1) * np.pi * t)

        err = weight * (x_target - x_est)
        return err

    def optimize(self, data, weight, w0, t0):
        ncenters = data.shape[0]
        self.w0 = w0
        self.t0 = t0
        if weight is None:
            weight = np.ones_like(t0)

        for iiter in range(self.maxiters):
            self.pp.append(self.calc_string(self.w0, self.t0, data))
            if iiter > 0:
                err = np.sum((self.pp[-1] - self.pp[-2]) ** 2) / ncenters
                print('{} -- {}'.format(iiter, err))
                if err < self.tol:
                    break
            else:
                print(iiter)
            # Optimize tk
            for ci in range(ncenters):
                self.t0[ci] = scipy.optimize.leastsq(self._optimize_dist, self.t0[ci], args=(data, self.w0, ci))[0]

            # Optimize wij
            for k in range(self.ndims):
                self.w0[k, :] = scipy.optimize.leastsq(self._optimize_w, self.w0[k, :], args=(data, self.t0, k, weight))[0]



from .core import ZMQWMError, ZMQWMTimeout, ZMQWMEnvironmentError, ZMQWorkerMissing, ZMQCore
from .node import ZMQNode
from .worker import ZMQWorker
from .work_manager import ZMQWorkManager

import atexit

__all__ = [
    'ZMQWMError',
    'ZMQWMTimeout',
    'ZMQWMEnvironmentError',
    'ZMQWorkerMissing',
    'ZMQCore',
    'ZMQNode',
    'ZMQWorker',
    'ZMQWorkManager',
]

atexit.register(ZMQCore.remove_ipc_endpoints)
del atexit



import logging
import os
import re

import h5py
import numpy as np

import matplotlib
from matplotlib import pyplot
from matplotlib.image import NonUniformImage

from westpa.tools import WESTMasterCommand, WESTSubcommand
from westpa.core import h5io, textio
from westpa.fasthist import normhistnd
from westpa.core.extloader import get_object

log = logging.getLogger('plothist')

# Suppress divide-by-zero in log
np.seterr(divide='ignore', invalid='ignore')


def sum_except_along(array, axes):
    '''Reduce the given array by addition over all axes except those listed in the scalar or
    iterable ``axes``'''

    try:
        iter(axes)
    except TypeError:
        axes = [axes]

    kept = set(axes)
    summed = list(set(range(array.ndim)) - kept)

    # Reorder axes so that the kept axes are first, and in the order they
    # were given
    array = np.transpose(array, list(axes) + summed).copy()

    # Now, the last len(summed) axes are summed over
    for _ in range(len(summed)):
        array = np.add.reduce(array, axis=-1)

    return array


class PlotHistBase(WESTSubcommand):
    def __init__(self, parent):
        super().__init__(parent)

        self.input_arg_group = None
        self.output_arg_group = None

        self.input_h5 = None
        self.opmode = None
        self.plotscale = None
        self.enerzero = None
        self.plotrange = None
        self.plottitle = None
        self.postprocess_function = None
        self.plot_contour = None

        # Iteration range for average/evolution
        self.avail_iter_start = None
        self.avail_iter_stop = None
        self.avail_iter_step = None
        self.iter_start = None
        self.iter_stop = None
        self.iter_step = None

        # Iteration for single point
        self.n_iter = None

        # An array of dicts describing what dimensions to work with and
        # what their ranges should be for the plots.
        self.dimensions = []

        self.plot_output_filename = None
        self.text_output_filename = None
        self.hdf5_output_filename = None

    def add_args(self, parser):
        igroup = self.input_arg_group = parser.add_argument_group('input options')
        igroup.add_argument('input', help='HDF5 file containing histogram data')
        igroup.add_argument(
            'firstdim',
            nargs='?',
            metavar='DIMENSION',
            help='''Plot for the given DIMENSION, specified as INT[:[LB,UB]:LABEL], where
                            INT is a zero-based integer identifying the dimension in the histogram,
                            LB and UB are lower and upper bounds for plotting, and LABEL is the label for
                            the plot axis. (Default: dimension 0, full range.)''',
        )

        ogroup = self.output_arg_group = parser.add_argument_group('output options')
        ogroup.add_argument(
            '-o',
            '--output',
            '--plot-output',
            dest='plot_output',
            default='hist.pdf',
            metavar='PLOT_OUTPUT',
            help='''Store plot as PLOT_OUTPUT. This may be set to an empty string
                            (e.g. --plot-output='') to suppress plotting entirely. The output
                            format is determined by filename extension (and thus defaults to PDF).
                            Default: "%(default)s".''',
        )
        ogroup.add_argument('--hdf5-output', help='''Store plot data in the HDF5 file HDF5_OUTPUT.''')
        ogroup.add_argument(
            '--plot-contour',
            dest='plot_contour',
            action='store_const',
            const=True,
            default=False,
            help='''Determines whether or not to superimpose a contour plot over the heatmap for 2D objects.''',
        )

        pgroup = parser.add_argument_group('plot options')
        pmgroup = pgroup.add_mutually_exclusive_group()
        pgroup.add_argument('--title', dest='title', help='Include TITLE as the top-of-graph title')
        pmgroup.add_argument(
            '--linear', dest='plotscale', action='store_const', const='linear', help='Plot the histogram on a linear scale.'
        )
        pmgroup.add_argument(
            '--energy',
            dest='plotscale',
            action='store_const',
            const='energy',
            help='Plot the histogram on an inverted natural log scale, corresponding to (free) energy (default).',
        )
        pmgroup.add_argument(
            '--zero-energy',
            dest='enerzero',
            metavar='E',
            default='min',
            help='Set the zero of energy to E, which may be a scalar, "min" or "max"',
        )
        pmgroup.add_argument(
            '--log10', dest='plotscale', action='store_const', const='log10', help='Plot the histogram on a base-10 log scale.'
        )
        pgroup.add_argument(
            '--range',
            help='''Plot histogram ordinates over the given RANGE, specified as "LB,UB",
                            where LB and UB are the lower and upper bounds, respectively. For 1-D plots,
                            this is the Y axis. For 2-D plots, this is the colorbar axis.
                            (Default: full range.)''',
        )
        pgroup.add_argument(
            '--postprocess-function',
            help='''Names a function (as in module.function) that will be called just prior
                                to saving the plot. The function will be called as ``postprocess(hist, midpoints, binbounds)``
                                where ``hist`` is the histogram that was plotted, ``midpoints`` is the bin midpoints for
                                each dimension, and ``binbounds`` is the bin boundaries for each dimension for 2-D plots,
                                or None otherwise. The plot must be modified in place using the pyplot stateful interface.''',
        )

        parser.set_defaults(plotscale='energy')

    def process_args(self, args):
        self.plotscale = args.plotscale
        self.input_h5 = h5py.File(args.input, 'r')
        self.plot_output_filename = args.plot_output
        self.hdf5_output_filename = args.hdf5_output
        self.plot_contour = args.plot_contour

        if args.title:
            self.plottitle = args.title

        if args.range:
            self.plotrange = self.parse_range(args.range)

        if args.firstdim:
            self.dimensions.append(self.parse_dimspec(args.firstdim))

        if not args.firstdim:
            self.dimensions.append({'idim': 0, 'label': 'dimension 0'})

        if args.enerzero:
            lenerzero = args.enerzero.lower()
            if lenerzero not in ('min', 'max'):
                try:
                    self.enerzero = float(args.enerzero)
                except ValueError:
                    raise ValueError('invalid energy zero point {!r}'.format(args.enerzero))
            else:
                self.enerzero = lenerzero
        else:
            self.enerzero = 'min'

        self.avail_iter_start, self.avail_iter_stop = h5io.get_iter_range(self.input_h5['histograms'])
        try:
            self.avail_iter_step = h5io.get_iter_step(self.input_h5['histograms'])
        except KeyError:
            self.avail_iter_step = 1
        log.info(
            'HDF5 file {!r} contains data for iterations {} -- {} with a step of {}'.format(
                args.input, self.avail_iter_start, self.avail_iter_stop, self.avail_iter_step
            )
        )
        if args.postprocess_function:
            self.postprocess_function = get_object(args.postprocess_function, path=['.'])

    def parse_dimspec(self, dimspec):
        dimdata = {}
        match = re.match(r'([0-9]+)(?::(?:([^,]+),([^:,]+))?(?::(.*))?)?', dimspec)
        if not match:
            raise ValueError('invalid dimension specification {!r}'.format(dimspec))

        (idim_txt, lb_txt, ub_txt, label) = match.groups()
        try:
            dimdata['idim'] = int(idim_txt)
            if lb_txt:
                dimdata['lb'] = float(lb_txt)
            if ub_txt:
                dimdata['ub'] = float(ub_txt)
            if label:
                dimdata['label'] = label
            else:
                dimdata['label'] = 'dimension {}'.format(dimdata['idim'])
        except ValueError as e:
            raise ValueError('invalid dimension specification {!r}: {!r}'.format(dimspec, e))
        return dimdata

    def parse_range(self, rangespec):
        try:
            (lbt, ubt) = rangespec.split(',')
            return float(lbt), float(ubt)
        except (ValueError, TypeError) as e:
            raise ValueError('invalid range specification {!r}: {!r}'.format(rangespec, e))

    def _ener_zero(self, hist):
        hist = -np.log(hist)
        if self.enerzero == 'min':
            np.subtract(hist, hist.min(), out=hist, casting="unsafe")
        elif self.enerzero == 'max':
            np.subtract(hist, hist.max(), out=hist, casting="unsafe")
        else:
            np.subtract(hist, self.enerzero, out=hist, casting="unsafe")
        return hist


class PlotSupports2D(PlotHistBase):
    def __init__(self, parent):
        super().__init__(parent)

    def add_args(self, parser):
        self.input_arg_group.add_argument(
            'seconddim',
            nargs='?',
            metavar='ADDTLDIM',
            help='''For instantaneous/average plots, plot along the given additional
                                          dimension, producing a color map.''',
        )
        self.output_arg_group.add_argument(
            '--text-output',
            help='''Store plot data in a text format at TEXT_OUTPUT. This option is
                                           only valid for 1-D histograms. (Default: no text output.)''',
        )

    def process_args(self, args):
        self.text_output_filename = args.text_output
        if args.seconddim is not None:
            self.dimensions.append(self.parse_dimspec(args.seconddim))

    def _do_1d_output(self, hist, idim, midpoints):
        enehist = self._ener_zero(hist)
        log10hist = np.log10(hist)

        if self.hdf5_output_filename:
            with h5py.File(self.hdf5_output_filename, 'w') as output_h5:
                h5io.stamp_creator_data(output_h5)
                output_h5.attrs['source_data'] = os.path.abspath(self.input_h5.filename)
                output_h5.attrs['source_dimension'] = idim
                output_h5['midpoints'] = midpoints
                output_h5['histogram'] = hist

        if self.text_output_filename:
            with textio.NumericTextOutputFormatter(self.text_output_filename) as output_file:
                output_file.write_header('source data: {} dimension {}'.format(os.path.abspath(self.input_h5.filename), idim))
                output_file.write_header('column 0: midpoint of bin')
                output_file.write_header('column 1: probability in bin')
                output_file.write_header('column 2: -ln P')
                output_file.write_header('column 3: log10 P')
                np.savetxt(output_file, np.column_stack([midpoints, hist, enehist, log10hist]))

        if self.plot_output_filename:
            if self.plotscale == 'energy':
                plothist = enehist
                label = r'$-\ln\,P(x)$'
            elif self.plotscale == 'log10':
                plothist = log10hist
                label = r'$\log_{10}\ P(x)$'
            else:
                plothist = hist
                label = r'$P(x)$'
            pyplot.figure()
            pyplot.plot(midpoints, plothist)
            pyplot.xlim(self.dimensions[0].get('lb'), self.dimensions[0].get('ub'))
            if self.plotrange:
                pyplot.ylim(*self.plotrange)
            pyplot.xlabel(self.dimensions[0]['label'])
            pyplot.ylabel(label)
            if self.plottitle:
                pyplot.title(self.plottitle)
            if self.postprocess_function:
                self.postprocess_function(plothist, midpoints, None)
            pyplot.savefig(self.plot_output_filename)

    def _do_2d_output(self, hist, idims, midpoints, binbounds):
        enehist = self._ener_zero(hist)
        log10hist = np.log10(hist)

        if self.hdf5_output_filename:
            with h5py.File(self.hdf5_output_filename, 'w') as output_h5:
                h5io.stamp_creator_data(output_h5)
                output_h5.attrs['source_data'] = os.path.abspath(self.input_h5.filename)
                output_h5.attrs['source_dimensions'] = np.array(idims, np.min_scalar_type(max(idims)))
                output_h5.attrs['source_dimension_labels'] = np.array([dim['label'] for dim in self.dimensions])
                for idim in idims:
                    output_h5['midpoints_{}'.format(idim)] = midpoints[idim]
                output_h5['histogram'] = hist

        if self.plot_output_filename:
            if self.plotscale == 'energy':
                plothist = enehist
                label = r'$-\ln\,P(x)$'
            elif self.plotscale == 'log10':
                plothist = log10hist
                label = r'$\log_{10}\ P(\vec{x})$'
            else:
                plothist = hist
                plothist[~np.isfinite(plothist)] = np.nan
                label = r'$P(\vec{x})$'

            try:
                vmin, vmax = self.plotrange
            except TypeError:
                vmin, vmax = None, None

            pyplot.figure()
            # Transpose input so that axis 0 is displayed as x and axis 1 is displayed as y
            #            pyplot.imshow(plothist.T, interpolation='nearest', aspect='auto',
            #                          extent=(midpoints[0][0], midpoints[0][-1], midpoints[1][0], midpoints[1][-1]),
            #                          origin='lower', vmin=vmin, vmax=vmax)

            # The following reproduces the former calls to imshow and colorbar
            norm = matplotlib.colors.Normalize(vmin=vmin, vmax=vmax)
            ax = pyplot.gca()
            nui = NonUniformImage(
                ax, extent=(midpoints[0][0], midpoints[0][-1], midpoints[1][0], midpoints[1][-1]), origin='lower', norm=norm
            )
            nui.set_data(midpoints[0], midpoints[1], plothist.T)
            ax.add_image(nui)
            ax.set_xlim(midpoints[0][0], midpoints[0][-1])
            ax.set_ylim(midpoints[1][0], midpoints[1][-1])
            cb = pyplot.colorbar(nui)
            cb.set_label(label)

            pyplot.xlabel(self.dimensions[0]['label'])
            pyplot.xlim(self.dimensions[0].get('lb'), self.dimensions[0].get('ub'))
            pyplot.ylabel(self.dimensions[1]['label'])
            pyplot.ylim(self.dimensions[1].get('lb'), self.dimensions[1].get('ub'))
            if self.plottitle:
                pyplot.title(self.plottitle)
            if self.postprocess_function:
                self.postprocess_function(plothist, midpoints, binbounds)
            if self.plot_contour:
                pyplot.contour(midpoints[0], midpoints[1], plothist.T)
            pyplot.savefig(self.plot_output_filename)


class InstantPlotHist(PlotSupports2D):
    subcommand = 'instant'
    help_text = 'plot probability distribution for a single WE iteration'
    description = '''\
Plot a probability distribution for a single WE iteration. The probability
distribution must have been previously extracted with ``w_pdist`` (or, at
least, must be compatible with the output format of ``w_pdist``; see
``w_pdist --help`` for more information).
'''

    def add_args(self, parser):
        self.input_arg_group.add_argument(
            '--iter',
            metavar='N_ITER',
            dest='n_iter',
            type=int,
            help='''Plot distribution for iteration N_ITER
                                          (default: last completed iteration).''',
        )

    def process_args(self, args):
        if args.n_iter:
            self.n_iter = min(args.n_iter, self.avail_iter_stop - 1)
        else:
            self.n_iter = self.avail_iter_stop - 1

    def do_instant_plot_1d(self):
        '''Plot the histogram for iteration self.n_iter'''

        idim = self.dimensions[0]['idim']
        n_iters = self.input_h5['n_iter'][...]
        iiter = np.searchsorted(n_iters, self.n_iter)
        binbounds = self.input_h5['binbounds_{}'.format(idim)][...]
        midpoints = self.input_h5['midpoints_{}'.format(idim)][...]
        hist = self.input_h5['histograms'][iiter]

        # Average over other dimensions
        hist = sum_except_along(hist, idim)
        normhistnd(hist, [binbounds])
        self._do_1d_output(hist, idim, midpoints)

    def do_instant_plot_2d(self):
        '''Plot the histogram for iteration self.n_iter'''

        idim0 = self.dimensions[0]['idim']
        idim1 = self.dimensions[1]['idim']

        n_iters = self.input_h5['n_iter'][...]
        iiter = np.searchsorted(n_iters, self.n_iter)
        binbounds_0 = self.input_h5['binbounds_{}'.format(idim0)][...]
        midpoints_0 = self.input_h5['midpoints_{}'.format(idim0)][...]
        binbounds_1 = self.input_h5['binbounds_{}'.format(idim1)][...]
        midpoints_1 = self.input_h5['midpoints_{}'.format(idim1)][...]

        hist = self.input_h5['histograms'][iiter]

        # Average over other dimensions
        hist = sum_except_along(hist, [idim0, idim1])
        normhistnd(hist, [binbounds_0, binbounds_1])
        self._do_2d_output(hist, [idim0, idim1], [midpoints_0, midpoints_1], [binbounds_0, binbounds_1])

    def go(self):
        if len(self.dimensions) == 2:
            self.do_instant_plot_2d()
        else:
            self.do_instant_plot_1d()


class AveragePlotHist(PlotSupports2D):
    subcommand = 'average'
    help_text = 'plot average of a probability distribution over a WE simulation'
    description = '''\
Plot a probability distribution averaged over multiple iterations. The
probability distribution must have been previously extracted with ``w_pdist``
(or, at least, must be compatible with the output format of ``w_pdist``; see
``w_pdist --help`` for more information).
'''

    def add_args(self, parser):
        igroup = self.input_arg_group
        igroup.add_argument(
            '--first-iter',
            dest='first_iter',
            type=int,
            metavar='N_ITER',
            default=1,
            help='''Begin averaging at iteration N_ITER (default: %(default)d).''',
        )
        igroup.add_argument(
            '--last-iter',
            dest='last_iter',
            type=int,
            metavar='N_ITER',
            help='''Conclude averaging with N_ITER, inclusive (default: last completed iteration).''',
        )

    def process_args(self, args):
        if args.first_iter:
            self.iter_start = max(args.first_iter, self.avail_iter_start)
        else:
            self.iter_start = self.avail_iter_start

        if args.last_iter:
            self.iter_stop = min(args.last_iter + 1, self.avail_iter_stop)
        else:
            self.iter_stop = self.avail_iter_stop

    def do_average_plot_1d(self):
        '''Plot the average histogram for iterations self.iter_start to self.iter_stop'''

        idim = self.dimensions[0]['idim']
        n_iters = self.input_h5['n_iter'][...]
        iiter_start = np.searchsorted(n_iters, self.iter_start)
        iiter_stop = np.searchsorted(n_iters, self.iter_stop)
        binbounds = self.input_h5['binbounds_{}'.format(idim)][...]
        midpoints = self.input_h5['midpoints_{}'.format(idim)][...]
        # hist = self.input_h5['histograms'][iiter_start:iiter_stop]

        for iiter in range(iiter_start, iiter_stop):
            iter_hist = sum_except_along(self.input_h5['histograms'][iiter], idim)
            if iiter == iiter_start:
                hist = iter_hist
            else:
                hist += iter_hist
            del iter_hist

        normhistnd(hist, [binbounds])
        self._do_1d_output(hist, idim, midpoints)

    def do_average_plot_2d(self):
        '''Plot the histogram for iteration self.n_iter'''

        idim0 = self.dimensions[0]['idim']
        idim1 = self.dimensions[1]['idim']

        n_iters = self.input_h5['n_iter'][...]
        iiter_start = np.searchsorted(n_iters, self.iter_start)
        iiter_stop = np.searchsorted(n_iters, self.iter_stop)

        binbounds_0 = self.input_h5['binbounds_{}'.format(idim0)][...]
        midpoints_0 = self.input_h5['midpoints_{}'.format(idim0)][...]
        binbounds_1 = self.input_h5['binbounds_{}'.format(idim1)][...]
        midpoints_1 = self.input_h5['midpoints_{}'.format(idim1)][...]

        for iiter in range(iiter_start, iiter_stop):
            iter_hist = sum_except_along(self.input_h5['histograms'][iiter], [idim0, idim1])
            if iiter == iiter_start:
                hist = iter_hist
            else:
                hist += iter_hist

        normhistnd(hist, [binbounds_0, binbounds_1])
        self._do_2d_output(hist, [idim0, idim1], [midpoints_0, midpoints_1], [binbounds_0, binbounds_1])

    def go(self):
        if len(self.dimensions) == 2:
            self.do_average_plot_2d()
        else:
            self.do_average_plot_1d()


class EvolutionPlotHist(PlotHistBase):
    subcommand = 'evolution'
    help_text = 'plot evolution of a probability distribution over the course of a WE simulation'
    description = '''\
Plot a probability distribution as it evolves over iterations. The
probability distribution must have been previously extracted with ``w_pdist``
(or, at least, must be compatible with the output format of ``w_pdist``; see
``w_pdist --help`` for more information).
'''

    def add_args(self, parser):
        igroup = self.input_arg_group
        igroup.add_argument(
            '--first-iter',
            dest='first_iter',
            type=int,
            metavar='N_ITER',
            default=1,
            help='''Begin analysis at iteration N_ITER (default: %(default)d).''',
        )
        igroup.add_argument(
            '--last-iter',
            dest='last_iter',
            type=int,
            metavar='N_ITER',
            help='''Conclude analysis with N_ITER, inclusive (default: last completed iteration).''',
        )
        igroup.add_argument(
            '--step-iter', dest='step_iter', type=int, metavar='STEP', help='''Average in blocks of STEP iterations.'''
        )

    def process_args(self, args):
        if args.first_iter:
            self.iter_start = max(args.first_iter, self.avail_iter_start)
        else:
            self.iter_start = self.avail_iter_start

        if args.last_iter:
            self.iter_stop = min(args.last_iter + 1, self.avail_iter_stop)
        else:
            self.iter_stop = self.avail_iter_stop

        if args.step_iter:
            self.iter_step = max(args.step_iter, self.avail_iter_step)
        else:
            self.iter_step = self.avail_iter_step
        log.info('using data for iterations {} -- {} with a step of {}'.format(self.iter_start, self.iter_stop, self.iter_step))

    def go(self):
        '''Plot the evolution of the histogram for iterations self.iter_start to self.iter_stop'''

        idim = self.dimensions[0]['idim']
        n_iters = self.input_h5['n_iter'][...]
        iiter_start = np.searchsorted(n_iters, self.iter_start)
        iiter_stop = np.searchsorted(n_iters, self.iter_stop)
        binbounds = self.input_h5['binbounds_{}'.format(idim)][...]
        midpoints = self.input_h5['midpoints_{}'.format(idim)][...]
        hists_ds = self.input_h5['histograms']

        itercount = self.iter_stop - self.iter_start

        # We always round down, so that we don't have a dangling partial block at the end
        nblocks = itercount // self.iter_step

        block_iters = np.empty((nblocks, 2), dtype=n_iters.dtype)
        blocked_hists = np.zeros((nblocks, hists_ds.shape[1 + idim]), dtype=hists_ds.dtype)

        for iblock, istart in enumerate(range(iiter_start, iiter_start + nblocks * self.iter_step, self.iter_step)):
            istop = min(istart + self.iter_step, iiter_stop)
            histslice = hists_ds[istart:istop]

            # Sum over time
            histslice = np.add.reduce(histslice, axis=0)

            # Sum over other dimensions
            blocked_hists[iblock] = sum_except_along(histslice, idim)

            # Normalize
            normhistnd(blocked_hists[iblock], [binbounds])

            block_iters[iblock, 0] = n_iters[istart]
            block_iters[iblock, 1] = n_iters[istop - 1] + 1

        # enehists = -np.log(blocked_hists)
        enehists = self._ener_zero(blocked_hists)
        log10hists = np.log10(blocked_hists)

        if self.hdf5_output_filename:
            with h5py.File(self.hdf5_output_filename, 'w') as output_h5:
                h5io.stamp_creator_data(output_h5)
                output_h5.attrs['source_data'] = os.path.abspath(self.input_h5.filename)
                output_h5.attrs['source_dimension'] = idim
                output_h5['midpoints'] = midpoints
                output_h5['histograms'] = blocked_hists
                output_h5['n_iter'] = block_iters

        if self.plot_output_filename:
            if self.plotscale == 'energy':
                plothist = enehists
                label = r'$-\ln\,P(x)$'
            elif self.plotscale == 'log10':
                plothist = log10hists
                label = r'$\log_{10}\ P(x)$'
            else:
                plothist = blocked_hists
                label = r'$P(x)$'

            try:
                vmin, vmax = self.plotrange
            except TypeError:
                vmin, vmax = None, None

            pyplot.figure()
            norm = matplotlib.colors.Normalize(vmin=vmin, vmax=vmax)
            ax = pyplot.gca()
            nui = NonUniformImage(
                ax, extent=(midpoints[0], midpoints[-1], block_iters[0, -1], block_iters[-1, -1]), origin='lower', norm=norm
            )

            # not sure why plothist works but plothist.T doesn't, and the opposite is true
            # for _do_2d_output
            nui.set_data(midpoints, block_iters[:, -1], plothist)
            ax.add_image(nui)
            ax.set_xlim(midpoints[0], midpoints[-1])
            ax.set_ylim(block_iters[0, -1], block_iters[-1, -1])
            cb = pyplot.colorbar(nui)
            cb.set_label(label)
            pyplot.xlabel(self.dimensions[0]['label'])
            pyplot.xlim(self.dimensions[0].get('lb'), self.dimensions[0].get('ub'))
            pyplot.ylabel('WE Iteration')
            if self.plottitle:
                pyplot.title(self.plottitle)
            if self.postprocess_function:
                self.postprocess_function(plothist, midpoints, binbounds)
            pyplot.savefig(self.plot_output_filename)


class PlotHistTool(WESTMasterCommand):
    prog = 'plothist'
    subparsers_title = 'plotting modes'
    subcommands = [InstantPlotHist, AveragePlotHist, EvolutionPlotHist]
    description = '''\
Plot probability density functions (histograms) generated by w_pdist or other
programs conforming to the same output format. This program operates in one of
three modes:

  instant
    Plot 1-D and 2-D histograms for an individual iteration. See
    ``plothist instant --help`` for more information.

  average
    Plot 1-D and 2-D histograms, averaged over several iterations. See
    ``plothist average --help`` for more information.

  evolution
    Plot the time evolution 1-D histograms as waterfall (heat map) plots.
    See ``plothist evolution --help`` for more information.

This program takes the output of ``w_pdist`` as input (see ``w_pdist --help``
for more information), and can generate any kind of graphical output that
matplotlib supports.


------------------------------------------------------------------------------
Command-line options
------------------------------------------------------------------------------
'''


def entry_point():
    PlotHistTool().main()


if __name__ == '__main__':
    entry_point()



import re

import numpy as np

import westpa
from .core import WESTToolComponent
from westpa.core.data_manager import seg_id_dtype

re_split_segspec = re.compile(r'\s*:\s*|\s+')


class SegmentSelection:
    def __init__(self, iterable=None):
        '''Initialize this segment selection from an iterable of (n_iter,seg_id) pairs.'''

        self._segments = set()
        self._segs_by_iter = {}
        self._start_iter = None
        self._stop_iter = None

        if iterable is not None:
            add = self.add
            for pair in iterable:
                add(tuple(pair))

    def __len__(self):
        return len(self._segments)

    def __contains__(self, pair):
        return tuple(pair) in self._segments

    def add(self, pair):

        (n_iter, seg_id) = int(pair[0]), int(pair[1])
        self._segments.add((n_iter, seg_id))
        self._segs_by_iter.setdefault(n_iter, set()).add(seg_id)
        if self._start_iter is None:
            self._start_iter = n_iter
        else:
            self._start_iter = min(self._start_iter, n_iter)
        if self._stop_iter is None:
            self._stop_iter = n_iter + 1
        else:
            self._stop_iter = max(self._stop_iter, n_iter + 1)

    def from_iter(self, n_iter):
        return self._segs_by_iter.get(n_iter, set())

    @property
    def start_iter(self):
        return self._start_iter

    @property
    def stop_iter(self):
        return self._stop_iter

    @classmethod
    def from_text(cls, filename):
        segsel = cls()

        segfile = open(filename, 'rt')
        for line in segfile:
            wline = line.strip()
            if wline[0] == '#' or wline == '':
                continue

            fields = re_split_segspec.split(wline)
            if len(fields) != 2:
                raise ValueError('malformed segment selection {!r}'.format(line))

            try:
                n_iter, seg_id = list(map(int, fields))
            except ValueError:
                raise ValueError('malformed segment selection {!r}'.format(line))

            segsel.add((n_iter, seg_id))
        return segsel


class AllSegmentSelection(SegmentSelection):
    def __init__(self, start_iter=None, stop_iter=None, data_manager=None):
        self.data_manager = data_manager or westpa.rc.get_data_manager()
        self._start_iter = start_iter or 1
        self._stop_iter = stop_iter or self.data_manager.current_iteration
        self._segcount_by_iter = {}

    def _count_from_iter(self, n_iter):
        try:
            segcount = self._segcount_by_iter[n_iter]
        except KeyError:
            segcount = self.data_manager.get_iter_group(n_iter)['seg_index'].shape[0]
            self._segcount_by_iter[n_iter] = segcount
        return segcount

    def add(self, pair):
        raise TypeError('cannot add segments to an AllSegmentSelection')

    def from_iter(self, n_iter):
        return np.arange(0, self._count_from_iter(n_iter), dtype=seg_id_dtype)

    def __len__(self):
        segcount = 0
        count_from_iter = self._count_from_iter
        for n_iter in range(self._start_iter, self._stop_iter):
            segcount += count_from_iter(n_iter)
        return segcount

    def __contains__(self, pair):
        (n_iter, seg_id) = pair
        return n_iter >= self._start_iter and n_iter < self._stop_iter and seg_id < self._count_from_iter(n_iter)


class SegSelector(WESTToolComponent):
    def __init__(self):
        super().__init__()
        self.segment_selection = None
        self.segsel_filename = None

    def add_args(self, parser):
        group = parser.add_argument_group('WEST input data options')
        sgroup = group.add_mutually_exclusive_group()
        sgroup.add_argument(
            '--segments-from',
            metavar='SEGLIST_FILE',
            help='''Include only segments listed in SEGLIST_FILE, as generated by w_select
                            (default: include all segments.''',
        )

    def process_args(self, args):
        if args.segments_from:
            self.segsel_filename = args.segments_from
            self.segment_selection = self.parse_segsel_file(self.segsel_filename)
        else:
            self.segment_selection = AllSegmentSelection()

    def parse_segsel_file(self, filename):
        segsel = self.segment_selection = SegmentSelection()

        segfile = open(filename, 'rt')
        for line in segfile:
            wline = line.strip()
            if wline[0] == '#' or wline == '':
                continue

            fields = re_split_segspec.split(wline)
            if len(fields) != 2:
                raise ValueError('malformed segment selection {!r}'.format(line))

            try:
                n_iter, seg_id = list(map(int, fields))
            except ValueError:
                raise ValueError('malformed segment selection {!r}'.format(line))

            segsel.add((n_iter, seg_id))
        return segsel



import numpy as np
from westpa.core.binning import FuncBinMapper
from westpa.core.extloader import get_object
import logging

log = logging.getLogger(__name__)


def map_binless(coords, mask, output, *args, **kwargs):
    '''Adaptively groups walkers according to a user-defined grouping function
    that is defined externally. Very general implementation but limited to
    only a two dimensional progress coordinate (for now).'''

    n_groups = kwargs.get("n_groups")
    n_dims = kwargs.get("n_dims")
    group_function = get_object(kwargs.get("group_function"))
    log.debug(f'binless arguments: {kwargs}')
    try:
        group_function_kwargs = kwargs.get('group_function_kwargs')['group_arguments']
    except KeyError:
        group_function_kwargs = {}
    ndim = n_dims

    if not np.any(mask):
        return output

    allcoords = np.copy(coords)
    allmask = np.copy(mask)

    isfinal = None
    splitting = False

    # the segments should be sent in by the driver as half initial segments and half final segments
    # allcoords contains all segments
    # coords should contain ONLY final segments
    if coords.shape[1] > ndim:
        if coords.shape[1] > ndim + 1:
            isfinal = allcoords[:, ndim + 1].astype(np.bool_)
        else:
            isfinal = np.ones(coords.shape[0], dtype=np.bool_)
        coords = coords[isfinal, :ndim]
        mask = mask[isfinal]
        splitting = True

    # in case where there is no final segments but initial ones in range
    if not np.any(mask):
        coords = allcoords[:, :ndim]
        mask = allmask
        splitting = False

    # filter the list of coordinates (which contains coordinates outside of the binless region)
    # to obtain only the ones we want to cluster
    # this is done with all dimensions at once
    binless_coords = coords[mask]
    nsegs_binless = len(binless_coords)

    # we need to make sure that the number of segments in the binless region is greater than
    # the number of clusters we request
    # if only one segment in the binless region, assign it to a single cluster
    if nsegs_binless == 1:
        clusters = [0]

    # if there are more than one segment in the binless region but still the total is less than
    # our target number, adjust our target number to be the number of segments in the binless
    # region minus one
    elif nsegs_binless < n_groups:
        clusters = group_function(binless_coords, nsegs_binless, splitting, **group_function_kwargs)
    # if there are enough segments in the binless region, proceed as planned
    elif nsegs_binless >= n_groups:
        clusters = group_function(binless_coords, n_groups, splitting, **group_function_kwargs)

    # this is a good place to say this... output is a list which matches the length of allcoords
    # allcoords is a collection of all initial and final segment coords for that iteration
    # we first filtered those to only contain the final data points, since those are the ones we care
    # about clustering
    # we then filtered to only have the coords in the binless region, since, again, those are what we care about
    # we then assigned each to a cluster which is essentially a slitting index
    # all that's left is to find where each binless segment is in the output and insert the cluster index there
    for idx, val in enumerate(binless_coords):
        if ndim > 1:
            mask2 = np.logical_and(allcoords[:, 0] == val[0], allcoords[:, 1] == val[1])
        else:
            mask2 = allcoords[:, 0] == val[0]
        output[mask2] = clusters[idx]

    return output


class BinlessMapper(FuncBinMapper):
    '''Adaptively group walkers according to a user-defined grouping
    function that is defined externally.'''

    def __init__(self, ngroups, ndims, group_function, **group_function_kwargs):
        kwargs = dict(n_groups=ngroups, n_dims=ndims, group_function=group_function, group_function_kwargs=group_function_kwargs)
        n_total_groups = ngroups
        super().__init__(map_binless, n_total_groups, kwargs=kwargs)



import re

import h5py
import numpy as np

from westpa.tools import WESTTool, WESTDataReader
import westpa
from westpa.core import h5io

from westpa.core.segment import Segment
from westpa.core.states import InitialState
from westpa.core.data_manager import weight_dtype, n_iter_dtype, seg_id_dtype, utime_dtype


class Trace:
    '''A class representing a trace of a certain trajectory segment back to its origin.'''

    def __init__(self, summary, endpoint_type, basis_state, initial_state, data_manager=None):
        self.summary = summary
        self.endpoint_type = endpoint_type
        self.basis_state = basis_state
        self.initial_state = initial_state
        self.data_manager = data_manager or westpa.rc.get_data_manager()

        # A mapping from aux file names to open h5py.File objects, to minimize time

        self._auxfiles = {}

    def __len__(self):
        try:
            return len(self.summary)
        except TypeError:
            return 0

    def __getitem__(self, sl):
        return self.summary[sl]

    def __iter__(self):
        return iter(self.summary)

    @classmethod
    def from_data_manager(cls, n_iter, seg_id, data_manager=None):
        '''Construct and return a trajectory trace whose last segment is identified
        by ``seg_id`` in the iteration number ``n_iter``.'''

        data_manager = data_manager or westpa.rc.get_data_manager()

        # These values are used later on
        endpoint_type = None
        pcoord_dtype = None
        pcoord_pt_shape = None

        seginfo = []
        parent_id = seg_id

        while n_iter > 0 and parent_id >= 0:
            seg_id = parent_id
            iter_group = data_manager.get_iter_group(n_iter)
            pcoord_ds = iter_group['pcoord']
            seg_index = iter_group['seg_index']
            n_segs = pcoord_ds.shape[0]
            pcoord_len = pcoord_ds.shape[1]

            assert seg_id < n_segs

            indexrow = seg_index[seg_id]
            final_pcoord = pcoord_ds[seg_id, pcoord_len - 1]
            weight = indexrow['weight']
            cputime = indexrow['cputime']
            walltime = indexrow['walltime']

            try:
                parent_id = int(indexrow['parent_id'])
            except IndexError:
                # old HDF5 version
                parent_id = int(iter_group['parents'][indexrow['parents_offset']])

            if endpoint_type is None:
                endpoint_type = indexrow['endpoint_type']
                pcoord_pt_shape = pcoord_ds.shape[2:]
                pcoord_dtype = pcoord_ds.dtype

            seginfo.append((n_iter, seg_id, weight, walltime, cputime, final_pcoord))

            del iter_group, pcoord_ds, seg_index
            n_iter -= 1

        # loop terminates with parent_id set to the identifier of the initial state,
        # seg_id set to the identifier of the first segment in the trajectory, and
        # n_iter set to one less than the iteration of the first segment
        first_iter = n_iter + 1
        first_seg_id = seg_id
        first_parent_id = parent_id

        # Initial segment (for fetching initial state)
        first_segment = Segment(n_iter=first_iter, seg_id=first_seg_id, parent_id=first_parent_id)

        seginfo.reverse()

        summary_dtype = np.dtype(
            [
                ('n_iter', n_iter_dtype),
                ('seg_id', seg_id_dtype),
                ('weight', weight_dtype),
                ('walltime', utime_dtype),
                ('cputime', utime_dtype),
                ('final_pcoord', pcoord_dtype, pcoord_pt_shape),
            ]
        )

        summary = np.array(seginfo, dtype=summary_dtype)

        try:
            initial_state = data_manager.get_segment_initial_states([first_segment], first_iter)[0]
        except KeyError:
            # old HDF5 version
            assert parent_id < 0
            istate_pcoord = data_manager.get_iter_group(first_iter)['pcoord'][first_seg_id, 0]
            istate_id = -(first_parent_id + 1)
            basis_state = None
            initial_state = InitialState(istate_id, None, iter_created=0, pcoord=istate_pcoord)

        else:
            basis_state = data_manager.get_basis_states(first_iter)[initial_state.basis_state_id]

        return cls(summary, endpoint_type, basis_state, initial_state, data_manager)

    def get_segment_data_slice(self, datafile, dsname, n_iter, seg_id, slice_=None, index_data=None, iter_prec=None):
        '''Return the data from the dataset named ``dsname`` within the given ``datafile`` (an open
        h5py.File object) for the given iteration and segment. By default, it is assumed that the
        dataset is stored in the iteration group for iteration ``n_iter``, but if ``index_data``
        is provided, it must be an iterable (preferably a simple array) of (n_iter,seg_id) pairs,
        and the index in the ``index_data`` iterable of the matching n_iter/seg_id pair is used as
        the index of the data to retrieve.

        If an optional ``slice_`` is provided, then the given slicing tuple is appended to that
        used to retrieve the segment-specific data (i.e. it can be used to pluck a subset of the
        data that would otherwise be returned).
        '''

        if slice_ is None:
            slice_ = np.s_[...]

        if index_data is not None:
            dataset = datafile[dsname]

            for i, (i_n_iter, i_seg_id) in enumerate(index_data):
                if (i_n_iter, i_seg_id) == (n_iter, seg_id):
                    break
            else:
                raise KeyError((n_iter, seg_id))

            itpl = (i,) + slice_
            return dataset[itpl]
        else:
            if not iter_prec:
                iter_prec = datafile.attrs.get('west_iter_prec', self.data_manager.default_iter_prec)
            igname_tail = 'iter_{:0{iter_prec:d}d}'.format(int(n_iter), iter_prec=int(iter_prec))
            try:
                iter_group = datafile['/iterations/' + igname_tail]
            except KeyError:
                iter_group = datafile[igname_tail]

            dataset = iter_group[dsname]
            itpl = (seg_id,) + slice_

            return dataset[itpl]

    def trace_timepoint_dataset(self, dsname, slice_=None, auxfile=None, index_ds=None):
        '''Return a trace along this trajectory over a dataset which is layed out as [seg_id][timepoint][...].
        Overlapping values at segment boundaries are accounted for.  Returns (data_trace, weight), where
        data_trace is a time series of the dataset along this trajectory, and weight is the corresponding
        trajectory weight at each time point.

        If ``auxfile`` is given, then load the dataset from the given HDF5 file, which must be
        layed out the same way as the main HDF5 file (e.g. iterations arranged as
        iterations/iter_*).

        If index_ds is given, instead of reading data per-iteration from iter_* groups, then the
        given index_ds is used as an index of n_iter,seg_id pairs into ``dsname``. In this case,
        the target data set need not exist on a per-iteration basis inside iter_* groups.

        If ``slice_`` is given, then *further* slice the data returned from the HDF5 dataset. This can
        minimize I/O if it is known (and specified) that only a subset of the data along the
        trajectory is needed.
        '''

        # Figure out where to look for the dataset
        if isinstance(auxfile, str):
            datafile = h5py.File(auxfile, 'r')
            close_datafile = True
        elif auxfile is not None:
            datafile = auxfile
            close_datafile = False
        else:
            datafile = self.data_manager.we_h5file
            close_datafile = False

        iter_prec = self.data_manager.iter_prec
        get_data_slice = self.get_segment_data_slice

        # Load the index if we use it
        if index_ds is not None:
            if isinstance(index_ds, str):
                index_ds = datafile[index_ds]
            index_data = index_ds[...]
        else:
            index_data = None

        # Be sure to retrieve the time series
        if not slice_:
            first_sl = np.index_exp[:, ...]
            other_sl = np.index_exp[1:, ...]
        else:
            first_sl = np.index_exp[:] + slice_
            other_sl = np.index_exp[1:] + slice_

        # Retrieve the first segment's data
        first_n_iter, first_seg_id = self.summary[0]['n_iter'], self.summary[0]['seg_id']
        first_iter_data = get_data_slice(datafile, dsname, first_n_iter, first_seg_id, first_sl, index_data, iter_prec)

        n_segs = len(self)
        n_points_per_seg = len(first_iter_data)

        length = n_points_per_seg + (n_segs - 1) * (n_points_per_seg - 1)
        tracedata = np.empty((length,) + first_iter_data.shape[1:], dtype=first_iter_data.dtype)
        traceweight = np.empty((length,), weight_dtype)

        # Store first segment data
        tracedata[0:n_points_per_seg] = first_iter_data
        traceweight[0:n_points_per_seg] = self.summary[0]['weight']
        del first_iter_data

        # Store remainder of data

        for iseg, summary_item in enumerate(self.summary[1:]):
            n_iter = summary_item['n_iter']
            seg_id = summary_item['seg_id']
            weight = summary_item['weight']

            offset = n_points_per_seg + iseg * (n_points_per_seg - 1)
            length = n_points_per_seg - 1
            seg_data = get_data_slice(datafile, dsname, n_iter, seg_id, other_sl, index_data, iter_prec)

            tracedata[offset : offset + length] = seg_data
            traceweight[offset : offset + length] = weight
            del seg_data

        if close_datafile:
            datafile.close()

        return tracedata, traceweight

    """
    # This is disabled until there is a real use for it; the following code is
    # outdated
    def trace_perseg_dataset(self, dsname):
        '''Return a trace along this trajectory over a dataset which is layed out as [seg_id][...].
        Returns (data_trace, weight), where  data_trace is a time series of the dataset along this
        trajectory, and weight is the corresponding trajectory weight at each time point.'''

        first_n_iter, first_seg_id = self.summary[0]['n_iter'], self.summary[0]['seg_id']
        first_iter_group = self.data_manager.get_iter_group(first_n_iter)
        first_iter_ds = first_iter_group[dsname]
        n_segs = len(self)
        tracedata = np.empty((n_segs,) + first_iter_ds.shape[1:], dtype=first_iter_ds.dtype)
        traceweight = np.empty((n_segs,), weight_dtype)
        tracedata[0] = first_iter_ds[first_seg_id]
        traceweight[0] = self.summary[0]['weight']
        for isegm1, summary_item in enumerate(self.summary[1:]):
            iseg = isegm1 + 1
            n_iter = summary_item['n_iter']
            seg_id = summary_item['seg_id']
            iter_group = self.data_manager.get_iter_group(n_iter)
            seg_data = iter_group[dsname][seg_id]
            tracedata[iseg] = seg_data
            traceweight[iseg] = summary_item['weight']
            del seg_data

        return tracedata, traceweight
    """


class WTraceTool(WESTTool):
    prog = 'w_trace'
    description = '''\
Trace individual WEST trajectories and emit (or calculate) quantities along the
trajectory.

Trajectories are specified as N_ITER:SEG_ID pairs. Each segment is traced back
to its initial point, and then various quantities (notably n_iter and seg_id)
are printed in order from initial point up until the given segment in the given
iteration.

Output is stored in several files, all named according to the pattern given by
the -o/--output-pattern parameter. The default output pattern is "traj_%d_%d",
where the printf-style format codes are replaced by the iteration number and
segment ID of the terminal segment of the trajectory being traced.

Individual datasets can be selected for writing using the -d/--dataset option
(which may be specified more than once). The simplest form is ``-d dsname``,
which causes data from dataset ``dsname`` along the trace to be stored to
HDF5.  The dataset is assumed to be stored on a per-iteration basis, with
the first dimension corresponding to seg_id and the second dimension
corresponding to time within the segment.  Further options are specified
as comma-separated key=value pairs after the data set name, as in

    -d dsname,alias=newname,index=idsname,file=otherfile.h5,slice=[100,...]

The following options for datasets are supported:

    alias=newname
        When writing this data to HDF5 or text files, use ``newname``
        instead of ``dsname`` to identify the dataset. This is mostly of
        use in conjunction with the ``slice`` option in order, e.g., to
        retrieve two different slices of a dataset and store then with
        different names for future use.

    index=idsname
        The dataset is not stored on a per-iteration basis for all
        segments, but instead is stored as a single dataset whose
        first dimension indexes n_iter/seg_id pairs. The index to
        these n_iter/seg_id pairs is ``idsname``.

    file=otherfile.h5
        Instead of reading data from the main WEST HDF5 file (usually
        ``west.h5``), read data from ``otherfile.h5``.

    slice=[100,...]
        Retrieve only the given slice from the dataset. This can be
        used to pick a subset of interest to minimize I/O.

-------------------------------------------------------------------------------
'''

    pcoord_formats = {
        'u8': '%20d',
        'i8': '%20d',
        'u4': '%10d',
        'i4': '%11d',
        'u2': '%5d',
        'i2': '%6d',
        'f4': '%14.7g',
        'f8': '%023.15g',
    }

    def __init__(self):
        super().__init__()

        self.data_reader = WESTDataReader()
        # self.h5storage = HDF5Storage()
        self.output_file = None
        self.output_pattern = None
        self.endpoints = None
        self.datasets = []

    # Interface for command-line tools
    def add_args(self, parser):
        self.data_reader.add_args(parser)
        # self.h5storage.add_args(parser)
        parser.add_argument(
            '-d',
            '--dataset',
            dest='datasets',
            # this breaks argparse (see http://bugs.python.org/issue11874)
            # metavar='DSNAME[,alias=ALIAS][,index=INDEX][,file=FILE][,slice=SLICE]',
            metavar='DSNAME',
            action='append',
            help='''Include the dataset named DSNAME in trace output. An extended form like
                            DSNAME[,alias=ALIAS][,index=INDEX][,file=FILE][,slice=SLICE] will
                            obtain the dataset from the given FILE instead of the main WEST HDF5 file,
                            slice it by SLICE, call it ALIAS in output, and/or access per-segment data by a n_iter,seg_id
                            INDEX instead of a seg_id indexed dataset in the group for n_iter.''',
        )
        parser.add_argument(
            'endpoints',
            metavar='N_ITER:SEG_ID',
            nargs='+',
            help='''Trace trajectory ending (or at least alive at) N_ITER:SEG_ID.''',
        )

        # tgroup = parser.add_argument_group('trace options')
        ogroup = parser.add_argument_group('output options')
        ogroup.add_argument(
            '--output-pattern',
            default='traj_%d_%d',
            help='''Write per-trajectory data to output files/HDF5 groups whose names begin with OUTPUT_PATTERN,
                                 which must contain two printf-style format flags which will be replaced with the iteration number
                                 and segment ID of the terminal segment of the trajectory being traced.
                                 (Default: %(default)s.)''',
        )
        ogroup.add_argument(
            '-o',
            '--output',
            default='trajs.h5',
            help='Store intermediate data and analysis results to OUTPUT (default: %(default)s).',
        )

    def process_args(self, args):
        self.data_reader.process_args(args)
        # self.h5storage.process_args(args)
        self.endpoints = [list(map(int, endpoint.split(':'))) for endpoint in args.endpoints]
        self.output_pattern = args.output_pattern

        for dsstr in args.datasets or []:
            self.datasets.append(self.parse_dataset_string(dsstr))

        # self.h5storage.open_analysis_h5file()
        self.output_file = h5py.File(args.output, 'a')

    def parse_dataset_string(self, dsstr):
        dsinfo = {}

        r = re.compile(r',(?=[^\]]*(?:\[|$))')
        fields = r.split(dsstr)

        dsinfo['dsname'] = fields[0]

        for field in (field.strip() for field in fields[1:]):
            k, v = field.split('=')
            k = k.lower()
            if k in ('alias', 'file', 'index'):
                dsinfo[k] = v
            elif k == 'slice':
                try:
                    dsinfo['slice'] = eval('np.index_exp' + v)
                except SyntaxError:
                    raise SyntaxError('invalid index expression {!r}'.format(v))
            else:
                raise ValueError('invalid dataset option {!r}'.format(k))

        return dsinfo

    def go(self):
        self.data_reader.open('r')

        # Create a new 'trajectories' group if this is the first trace
        try:
            trajs_group = h5io.create_hdf5_group(self.output_file, 'trajectories', replace=False, creating_program=self.prog)
        except ValueError:
            trajs_group = self.output_file['trajectories']

        for n_iter, seg_id in self.endpoints:
            trajname = self.output_pattern % (n_iter, seg_id)
            trajgroup = trajs_group.create_group(trajname)

            trace = Trace.from_data_manager(n_iter, seg_id, self.data_reader.data_manager)

            with open(trajname + '_trace.txt', 'wt') as trace_output:
                self.emit_trace_text(trace, trace_output)

            self.emit_trace_h5(trace, trajgroup)

            aux_h5files = {}
            for dsinfo in self.datasets:
                dsname = dsinfo['dsname']
                filename = dsinfo.get('file')
                if filename:
                    try:
                        aux_h5file = aux_h5files[filename]
                    except KeyError:
                        aux_h5file = aux_h5files[filename] = h5py.File(filename, 'r')
                else:
                    aux_h5file = None

                slice_ = dsinfo.get('slice')
                alias = dsinfo.get('alias', dsname)
                index = dsinfo.get('index')

                data, weights = trace.trace_timepoint_dataset(dsname, auxfile=aux_h5file, slice_=slice_, index_ds=index)

                # Save data to HDF5
                try:
                    del trajgroup[alias]
                except KeyError:
                    pass
                trajgroup[alias] = data

                # All weight vectors will be the same length, so only store in HDF5 once
                if not ('weights' in trajgroup and trajgroup['weights'].shape == weights.shape):
                    try:
                        del trajgroup['weights']
                    except KeyError:
                        pass
                    trajgroup['weights'] = weights

    def emit_trace_h5(self, trace, output_group):
        for dsname in ('basis_state', 'initial_state', 'segments'):
            try:
                del output_group[dsname]
            except KeyError:
                pass

        if trace.basis_state:
            output_group['basis_state'] = trace.basis_state.as_numpy_record()
        output_group['initial_state'] = trace.initial_state.as_numpy_record()
        output_group['segments'] = trace.summary

    def emit_trace_text(self, trace, output_file):
        '''Dump summary information about each segment in the given trace to the given output_file,
        which must be opened for writing in text mode.  Output columns are separated by at least
        one space.'''

        if not trace:
            return

        pcoord_ndim = trace[0]['final_pcoord'].shape[0]
        lastseg = trace[-1]
        len_n_iter = max(6, len(str(lastseg['n_iter'])))
        len_seg_id = max(6, max(len(str(seg_id)) for seg_id in trace['seg_id']))
        seg_pattern = (
            '    '.join(
                [
                    '{n_iter:{len_n_iter}d}',
                    '{seg_id:{len_seg_id}d}',
                    '{weight:22.17e}',
                    '{walltime:10.6g}',
                    '{cputime:10.6g}',
                    '{pcoord_str:s}',
                ]
            )
            + '\n'
        )

        output_file.write(
            '''\
# Trace of trajectory ending in n_iter:seg_id {n_iter:d}:{seg_id:d} (endpoint type {endpoint_type_text:s})
# column  0: iteration (0 => initial state)
# column  1: seg_id (or initial state ID)
# column  2: weight
# column  3: wallclock time (s)
# column  4: CPU time (s)
'''.format(
                n_iter=int(lastseg['n_iter']),
                seg_id=int(lastseg['seg_id']),
                endpoint_type_text=Segment.endpoint_type_names[trace.endpoint_type],
            )
        )

        if pcoord_ndim == 1:
            output_file.write(
                '''\
# column  5: final progress coordinate value
'''
            )
        else:
            fpcbegin = 5
            fpcend = fpcbegin + pcoord_ndim - 1
            output_file.write(
                '''\
# columns {fpcbegin:d} -- {fpcend:d}: final progress coordinate value
'''.format(
                    fpcbegin=fpcbegin, fpcend=fpcend
                )
            )

        pcoord_formats = self.pcoord_formats

        # Output row for initial state
        initial_state = trace.initial_state
        pcoord_str = '    '.join(pcoord_formats.get(pcfield.dtype.str[1:], '%s') % pcfield for pcfield in initial_state.pcoord)
        output_file.write(
            seg_pattern.format(
                n_iter=0,
                seg_id=initial_state.state_id,
                weight=0.0,
                walltime=0,
                cputime=0,
                pcoord_str=pcoord_str,
                len_n_iter=len_n_iter,
                len_seg_id=len_seg_id,
            )
        )

        # Output rows for segments
        for segment in trace:
            pcoord_str = '    '.join(
                pcoord_formats.get(pcfield.dtype.str[1:], '%s') % pcfield for pcfield in segment['final_pcoord']
            )
            output_file.write(
                seg_pattern.format(
                    n_iter=int(segment['n_iter']),
                    seg_id=int(segment['seg_id']),
                    weight=float(segment['weight']),
                    walltime=float(segment['walltime']),
                    cputime=float(segment['cputime']),
                    pcoord_str=pcoord_str,
                    len_n_iter=len_n_iter,
                    len_seg_id=len_seg_id,
                )
            )


def entry_point():
    WTraceTool().main()


if __name__ == '__main__':
    entry_point()



'''Miscellaneous routines to help with input and output of WEST-related data in text format'''

import getpass
import socket
import sys
import time


class NumericTextOutputFormatter:
    comment_string = '# '
    emit_header = True

    def __init__(self, output_file, mode='wt', emit_header=None):
        if hasattr(output_file, 'write'):
            self._file = output_file
        else:
            self._file = open(output_file, mode)

        self._header_written = False
        self._header_lines = []

        self.write_header('created by program: {}'.format(sys.argv[0] or 'unknown program'))
        self.write_header('created by user:    {}'.format(getpass.getuser()))
        self.write_header('created on host:    {}'.format(socket.gethostname()))
        self.write_header('created at:         {}'.format(time.strftime('%c', time.localtime(time.time()))))

        if emit_header is not None:
            self.emit_header = emit_header
        else:
            self.emit_header = self.__class__.emit_header

    def __getattr__(self, attr):
        return getattr(self._file, attr)

    def close(self):
        if not self._header_written:
            self._write_header()
        self._file.close()

    def write(self, str):
        if not self._header_written:
            self._write_header()
        self._file.write(str)

    def writelines(self, sequence):
        if not self._header_written:
            self._write_header()
        self._file.writelines(sequence)

    def write_comment(self, line):
        '''Writes a line beginning with the comment string'''
        self._file.write('{}{}\n'.format(self.comment_string, line))

    def write_header(self, line):
        '''Appends a line to those written when the file header is written. The
        appropriate comment string will be prepended, so ``line`` should not include
        a comment character.'''
        if self._header_written:
            raise EnvironmentError('cannot append header lines after header has been written')
        else:
            self._header_lines.append(line)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._file.close()
        return False

    def _write_header(self):
        if self.emit_header and not self._header_written:
            for line in self._header_lines:
                self.write_comment(line)
        self._header_written = True



import argparse
import io
import logging
import sys

import numpy as np

import westpa

from westpa.core.states import BasisState, TargetState

import westpa.work_managers as work_managers
from westpa.work_managers import make_work_manager


log = logging.getLogger('w_init')
EPS = np.finfo(np.float64).eps


def entry_point():

    parser = argparse.ArgumentParser(
        'w_init',
        description='''\
    Initialize a new WEST simulation, creating the WEST HDF5 file and preparing the first
    iteration's segments.

    Initial states are generated from one or more "basis states" which are specified either in a file specified with --bstates-from, or
    by one or more "--bstate" arguments. If neither --bstates-from nor at least one "--bstate" argument is provided, then a default
    basis state of probability one identified by the state ID zero and label "basis" will be created (a warning will be printed in this
    case, to remind you of this behavior, in case it is not what you wanted).

    Target states for (non-equilibrium) steady-state simulations are specified either in a file specified with --tstates-from, or
    by one or more --tstate arguments. If neither --tstates-from nor at least one --tstate argument is provided, then an equilibrium
    simulation (without any sinks) will be performed.
    ''',
    )
    westpa.rc.add_args(parser)
    parser.add_argument('--force', dest='force', action='store_true', help='Overwrite any existing simulation data')
    parser.add_argument(
        '--bstate-file',
        '--bstates-from',
        metavar='BSTATE_FILE',
        help='Read basis state names, probabilities, and (optionally) data references from BSTATE_FILE.',
    )
    parser.add_argument(
        '--bstate',
        action='append',
        dest='bstates',
        help='''Add the given basis state (specified as a string 'label,probability[,auxref]')
                        to the list of basis states (after those specified in --bstates-from, if any). This argument
                        may be specified more than once, in which case the given states are appended in the order
                        they are given on the command line.''',
    )
    parser.add_argument(
        '--tstate-file',
        '--tstates-from',
        metavar='TSTATE_FILE',
        help='''Read target state names and representative progress coordinates from TSTATE_FILE. WESTPA uses the
                        representative progress coordinate of a target state and converts the **entire** bin
                        containing that progress coordinate into a recycling sink.''',
    )
    parser.add_argument(
        '--tstate',
        action='append',
        dest='tstates',
        help='''Add the given target state (specified as a string 'label,pcoord0[,pcoord1[,...]]') to the
                        list of target states (after those specified in the file given by --tstates-from, if any).
                        This argument may be specified more than once, in which case the given states are appended
                        in the order they appear on the command line.''',
    )

    parser.add_argument(
        '--sstate-file',
        '--sstates-from',
        metavar='SSTATE_FILE',
        help='Read start state names, probabilities, and (optionally) data references from SSTATE_FILE.',
    )
    parser.add_argument(
        '--sstate',
        action='append',
        dest='sstates',
        help='''Add the given start state (specified as a string 'label,probability[,auxref]')
                        to the list of start states (after those specified in --sstates-from, if any). This argument
                        may be specified more than once, in which case the given states are appended in the order
                        they are given on the command line.''',
    )

    parser.add_argument(
        '--segs-per-state',
        type=int,
        metavar='N',
        default=1,
        help='''Initialize N segments from each basis state (default: %(default)s).''',
    )
    parser.add_argument(
        '--no-we',
        '--shotgun',
        dest='shotgun',
        action='store_true',
        help='''Do not run the weighted ensemble bin/split/merge algorithm on newly-created segments.''',
    )

    # TODO: Does this belong here or not? I like that it's parsing arguments, which is the purpose of entry_point.
    #   I don't necessarily like that it's setting state across different parts of the program.

    work_managers.environment.add_wm_args(parser)
    args = parser.parse_args()
    westpa.rc.process_args(args)
    work_managers.environment.process_wm_args(args)

    initialize(
        args.tstates,
        args.tstate_file,
        args.bstates,
        args.bstate_file,
        args.sstates,
        args.sstate_file,
        args.segs_per_state,
        args.shotgun,
    )


def initialize(tstates, tstate_file, bstates, bstate_file, sstates=None, sstate_file=None, segs_per_state=1, shotgun=False):
    """
    Initialize a WESTPA simulation.

    tstates : list of str

    tstate_file : str

    bstates : list of str

    bstate_file : str

    sstates : list of str

    sstate_file : str

    segs_per_state : int

    shotgun : bool
    """

    westpa.rc.work_manager = work_manager = make_work_manager()

    system = westpa.rc.get_system_driver()
    sim_manager = westpa.rc.get_sim_manager()

    data_manager = westpa.rc.get_data_manager()

    data_manager.system = system

    with work_manager:
        if work_manager.is_master:
            # Process target states
            target_states = []
            if tstate_file:
                target_states.extend(TargetState.states_from_file(tstate_file, system.pcoord_dtype))
            if tstates:
                tstates_strio = io.StringIO('\n'.join(tstates).replace(',', ' '))
                target_states.extend(TargetState.states_from_file(tstates_strio, system.pcoord_dtype))
                del tstates_strio

            # Process basis states
            basis_states = []
            if bstate_file:
                basis_states.extend(BasisState.states_from_file(bstate_file))
            if bstates:
                for bstate_str in bstates:
                    fields = bstate_str.split(',')
                    label = fields[0]
                    probability = float(fields[1])
                    try:
                        auxref = fields[2]
                    except IndexError:
                        auxref = None
                    basis_states.append(BasisState(label=label, probability=probability, auxref=auxref))

            # Process the list of start states, creating a BasisState from each
            start_states = []
            if sstate_file:
                start_states.extend(BasisState.states_from_file(sstate_file))
            if sstates:
                for sstate_str in sstates:
                    fields = sstate_str.split(',')
                    label = fields[0]
                    probability = float(fields[1])
                    try:
                        auxref = fields[2]
                    except IndexError:
                        auxref = None
                    start_states.append(BasisState(label=label, probability=probability, auxref=auxref))

            if not basis_states:
                log.error('At least one basis state is required')
                sys.exit(3)

            # Check that the total probability of basis states adds to one
            bstate_prob, sstate_prob = (
                sum(bstate.probability for bstate in basis_states),
                sum(sstate.probability for sstate in start_states),
            )
            # tprob = sum(bstate.probability for bstate in basis_states)
            tprob = bstate_prob + sstate_prob
            if abs(1.0 - tprob) > len(basis_states) * EPS:
                pscale = 1 / tprob
                log.warning(
                    'Basis state probabilities do not add to unity (basis: {:.2f}, start states: {:.2f}); rescaling by {:g}. If using start states, some rescaling is normal.'.format(
                        bstate_prob, sstate_prob, pscale
                    )
                )
                for bstate in basis_states:
                    bstate.probability *= pscale
                for sstate in start_states:
                    sstate.probability *= pscale

            # Prepare simulation
            sim_manager.initialize_simulation(
                basis_states=basis_states,
                target_states=target_states,
                start_states=start_states,
                segs_per_state=segs_per_state,
                suppress_we=shotgun,
            )
        else:
            work_manager.run()


if __name__ == '__main__':
    entry_point()
